import numpy as np

import wisdem.inputs as sch


class WindTurbineOntologyPython(object):
    def __init__(self, fname_input_wt, fname_input_modeling, fname_input_analysis):

        self.modeling_options = sch.load_modeling_yaml(fname_input_modeling)
        self.analysis_options = sch.load_analysis_yaml(fname_input_analysis)
        if fname_input_wt is None:
            self.wt_init = None
        else:
            self.wt_init = sch.load_geometry_yaml(fname_input_wt)
        self.set_run_flags()
        self.set_openmdao_vectors()
        self.set_opt_flags()

    def get_input_data(self):
        return self.wt_init, self.modeling_options, self.analysis_options

    def set_run_flags(self):
        # Create components flag struct
        self.modeling_options["flags"] = {}

        # Backwards compatibility
        modules = ["RotorSE", "DriveSE", "GeneratorSE", "TowerSE", "FixedBottomSE", "FloatingSE", "Loading", "BOS"]
        for m in modules:
            if m in self.modeling_options:
                self.modeling_options["WISDEM"][m].update(self.modeling_options[m])

        for k in ["blade", "hub", "nacelle", "tower", "monopile", "jacket", "floating_platform", "mooring", "RNA"]:
            self.modeling_options["flags"][k] = k in self.wt_init["components"]

        for k in ["assembly", "components", "airfoils", "materials", "control", "environment", "bos", "costs"]:
            self.modeling_options["flags"][k] = k in self.wt_init

        # Generator flag
        self.modeling_options["flags"]["generator"] = False
        if self.modeling_options["flags"]["nacelle"] and "generator" in self.wt_init["components"]["nacelle"]:
            self.modeling_options["flags"]["generator"] = True
            if not "GeneratorSE" in self.modeling_options["WISDEM"]:
                self.modeling_options["WISDEM"]["GeneratorSE"] = {}
            self.modeling_options["WISDEM"]["GeneratorSE"]["type"] = self.wt_init["components"]["nacelle"]["generator"][
                "generator_type"
            ].lower()

        # Offshore flags
        self.modeling_options["flags"]["floating"] = self.modeling_options["flags"]["floating_platform"]
        self.modeling_options["flags"]["offshore"] = (
            self.modeling_options["flags"]["floating"]
            or self.modeling_options["flags"]["monopile"]
            or self.modeling_options["flags"]["jacket"]
        )

        # Put in some logic about what needs to be in there
        flags = self.modeling_options["flags"]

        # Even if the block is in the inputs, the user can turn off via modeling options
        if flags["bos"]:
            flags["bos"] = self.modeling_options["WISDEM"]["BOS"]["flag"]
        if flags["blade"]:
            flags["blade"] = self.modeling_options["WISDEM"]["RotorSE"]["flag"]
        if flags["tower"]:
            flags["tower"] = self.modeling_options["WISDEM"]["TowerSE"]["flag"]
        if flags["monopile"]:
            flags["monopile"] = self.modeling_options["WISDEM"]["FixedBottomSE"]["flag"]
        if flags["jacket"]:
            flags["jacket"] = self.modeling_options["WISDEM"]["FixedBottomSE"]["flag"]
        if flags["hub"]:
            flags["hub"] = self.modeling_options["WISDEM"]["DriveSE"]["flag"]
        if flags["nacelle"]:
            flags["nacelle"] = self.modeling_options["WISDEM"]["DriveSE"]["flag"]
        if flags["generator"]:
            flags["generator"] = self.modeling_options["WISDEM"]["DriveSE"]["flag"]
        flags["hub"] = flags["nacelle"] = flags["hub"] or flags["nacelle"]  # Hub and nacelle have to go together

        # Blades and airfoils
        if flags["blade"] and not flags["airfoils"]:
            raise ValueError("Blades/rotor analysis is requested but no airfoils are found")
        if flags["airfoils"] and not flags["blade"]:
            print("WARNING: Airfoils provided but no blades/rotor found or RotorSE deactivated")

        # Blades, tower, monopile and environment
        if flags["blade"] and not flags["environment"]:
            raise ValueError("Blades/rotor analysis is requested but no environment input found")
        if flags["tower"] and not flags["environment"]:
            raise ValueError("Tower analysis is requested but no environment input found")
        if flags["monopile"] and not flags["environment"]:
            raise ValueError("Monopile analysis is requested but no environment input found")
        if flags["jacket"] and not flags["environment"]:
            raise ValueError("Jacket analysis is requested but no environment input found")
        if flags["jacket"] and flags["monopile"]:
            raise ValueError("Cannot specify both monopile and jacket support structures")
        if flags["floating_platform"] and not flags["environment"]:
            raise ValueError("Floating analysis is requested but no environment input found")
        if flags["environment"] and not (
            flags["blade"] or flags["tower"] or flags["monopile"] or flags["jacket"] or flags["floating_platform"]
        ):
            print("WARNING: Environment provided but no related component found found")

        # Floating/monopile
        if flags["floating_platform"] and (flags["monopile"] or flags["jacket"]):
            raise ValueError("Cannot have both floating and fixed-bottom components")

        # Water depth check
        if "water_depth" in self.wt_init["environment"]:
            if self.wt_init["environment"]["water_depth"] <= 0.0 and flags["offshore"]:
                raise ValueError("Water depth must be > 0 to do fixed-bottom or floating analysis")

    def set_openmdao_vectors(self):
        # Class instance to determine all the parameters used to initialize the openmdao arrays, i.e. number of airfoils, number of angles of attack, number of blade spanwise stations, etc
        # ==modeling_options = {}

        # Materials
        self.modeling_options["materials"] = {}
        self.modeling_options["materials"]["n_mat"] = len(self.wt_init["materials"])

        # Airfoils
        if self.modeling_options["flags"]["airfoils"]:
            self.modeling_options["WISDEM"]["RotorSE"]["n_af"] = len(self.wt_init["airfoils"])
            self.modeling_options["WISDEM"]["RotorSE"]["n_aoa"] = self.modeling_options["WISDEM"]["RotorSE"]["n_aoa"]
            if self.modeling_options["WISDEM"]["RotorSE"]["n_aoa"] / 4.0 == int(
                self.modeling_options["WISDEM"]["RotorSE"]["n_aoa"] / 4.0
            ):
                # One fourth of the angles of attack from -pi to -pi/6, half between -pi/6 to pi/6, and one fourth from pi/6 to pi
                self.modeling_options["WISDEM"]["RotorSE"]["aoa"] = np.unique(
                    np.hstack(
                        [
                            np.linspace(
                                -np.pi, -np.pi / 6.0, int(self.modeling_options["WISDEM"]["RotorSE"]["n_aoa"] / 4.0 + 1)
                            ),
                            np.linspace(
                                -np.pi / 6.0,
                                np.pi / 6.0,
                                int(self.modeling_options["WISDEM"]["RotorSE"]["n_aoa"] / 2.0),
                            ),
                            np.linspace(
                                np.pi / 6.0, np.pi, int(self.modeling_options["WISDEM"]["RotorSE"]["n_aoa"] / 4.0 + 1)
                            ),
                        ]
                    )
                )
            else:
                self.modeling_options["WISDEM"]["RotorSE"]["aoa"] = np.linspace(
                    -np.pi, np.pi, self.modeling_options["WISDEM"]["RotorSE"]["n_aoa"]
                )
                print(
                    "WARNING: If you like a grid of angles of attack more refined between +- 30 deg, please choose a n_aoa in the analysis option input file that is a multiple of 4. The current value of "
                    + str(self.modeling_options["WISDEM"]["RotorSE"]["n_aoa"])
                    + " is not a multiple of 4 and an equally spaced grid is adopted."
                )
            Re_all = []
            self.modeling_options["WISDEM"]["RotorSE"]["AFTabMod"] = 1
            for i in range(self.modeling_options["WISDEM"]["RotorSE"]["n_af"]):
                for j in range(len(self.wt_init["airfoils"][i]["polars"])):
                    Re_all.append(self.wt_init["airfoils"][i]["polars"][j]["re"])
                if len(self.wt_init["airfoils"][i]["polars"]) > 1:
                    self.modeling_options["WISDEM"]["RotorSE"]["AFTabMod"] = 2
            self.modeling_options["WISDEM"]["RotorSE"]["n_Re"] = len(np.unique(Re_all))
            self.modeling_options["WISDEM"]["RotorSE"]["n_tab"] = 1
            self.modeling_options["WISDEM"]["RotorSE"]["n_xy"] = self.modeling_options["WISDEM"]["RotorSE"]["n_xy"]
            self.modeling_options["WISDEM"]["RotorSE"]["af_used"] = self.wt_init["components"]["blade"][
                "outer_shape_bem"
            ]["airfoil_position"]["labels"]

        # Blade
        self.modeling_options["WISDEM"]["RotorSE"]["bjs"] = False
        if self.modeling_options["flags"]["blade"]:
            self.modeling_options["WISDEM"]["RotorSE"]["nd_span"] = np.linspace(
                0.0, 1.0, self.modeling_options["WISDEM"]["RotorSE"]["n_span"]
            )  # Equally spaced non-dimensional spanwise grid
            self.modeling_options["WISDEM"]["RotorSE"]["n_af_span"] = len(
                self.wt_init["components"]["blade"]["outer_shape_bem"]["airfoil_position"]["labels"]
            )  # This is the number of airfoils defined along blade span and it is often different than n_af, which is the number of airfoils defined in the airfoil database
            self.modeling_options["WISDEM"]["RotorSE"]["n_webs"] = len(
                self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["webs"]
            )
            self.modeling_options["WISDEM"]["RotorSE"]["n_layers"] = len(
                self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"]
            )
            self.modeling_options["WISDEM"]["RotorSE"]["lofted_output"] = False
            self.modeling_options["WISDEM"]["RotorSE"]["n_freq"] = 10  # Number of blade nat frequencies computed

            self.modeling_options["WISDEM"]["RotorSE"]["layer_name"] = self.modeling_options["WISDEM"]["RotorSE"][
                "n_layers"
            ] * [""]
            self.modeling_options["WISDEM"]["RotorSE"]["layer_mat"] = self.modeling_options["WISDEM"]["RotorSE"][
                "n_layers"
            ] * [""]
            for i in range(self.modeling_options["WISDEM"]["RotorSE"]["n_layers"]):
                self.modeling_options["WISDEM"]["RotorSE"]["layer_name"][i] = self.wt_init["components"]["blade"][
                    "internal_structure_2d_fem"
                ]["layers"][i]["name"]
                self.modeling_options["WISDEM"]["RotorSE"]["layer_mat"][i] = self.wt_init["components"]["blade"][
                    "internal_structure_2d_fem"
                ]["layers"][i]["material"]

            self.modeling_options["WISDEM"]["RotorSE"]["web_name"] = self.modeling_options["WISDEM"]["RotorSE"][
                "n_webs"
            ] * [""]
            for i in range(self.modeling_options["WISDEM"]["RotorSE"]["n_webs"]):
                self.modeling_options["WISDEM"]["RotorSE"]["web_name"][i] = self.wt_init["components"]["blade"][
                    "internal_structure_2d_fem"
                ]["webs"][i]["name"]

            # Distributed aerodynamic control devices along blade
            self.modeling_options["WISDEM"]["RotorSE"]["n_te_flaps"] = 0
            if "aerodynamic_control" in self.wt_init["components"]["blade"]:
                if "te_flaps" in self.wt_init["components"]["blade"]["aerodynamic_control"]:
                    self.modeling_options["WISDEM"]["RotorSE"]["n_te_flaps"] = len(
                        self.wt_init["components"]["blade"]["aerodynamic_control"]["te_flaps"]
                    )
                    self.modeling_options["WISDEM"]["RotorSE"]["n_tab"] = 3
                else:
                    raise RuntimeError(
                        "A distributed aerodynamic control device is provided in the yaml input file, but not supported by wisdem."
                    )

            joint_pos = self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["joint"]["position"]
            if joint_pos > 0.0:
                self.modeling_options["WISDEM"]["RotorSE"]["bjs"] = True
                # Adjust grid to have grid point at join location
                closest_grid_pt = np.argmin(abs(self.modeling_options["WISDEM"]["RotorSE"]["nd_span"] - joint_pos))
                self.modeling_options["WISDEM"]["RotorSE"]["nd_span"][closest_grid_pt] = joint_pos
                self.modeling_options["WISDEM"]["RotorSE"]["id_joint_position"] = closest_grid_pt

        # Drivetrain
        if self.modeling_options["flags"]["nacelle"]:
            self.modeling_options["WISDEM"]["DriveSE"]["direct"] = self.wt_init["assembly"]["drivetrain"].lower() in [
                "direct",
                "direct_drive",
                "pm_direct_drive",
            ]

        # Tower
        if self.modeling_options["flags"]["tower"]:
            self.modeling_options["WISDEM"]["TowerSE"]["n_height"] = len(
                self.wt_init["components"]["tower"]["outer_shape_bem"]["outer_diameter"]["grid"]
            )
            self.modeling_options["WISDEM"]["TowerSE"]["n_layers"] = len(
                self.wt_init["components"]["tower"]["internal_structure_2d_fem"]["layers"]
            )
            self.modeling_options["WISDEM"]["TowerSE"]["n_height_tower"] = self.modeling_options["WISDEM"]["TowerSE"][
                "n_height"
            ]
            self.modeling_options["WISDEM"]["TowerSE"]["n_layers_tower"] = self.modeling_options["WISDEM"]["TowerSE"][
                "n_layers"
            ]

        # Monopile
        if self.modeling_options["flags"]["monopile"]:
            self.modeling_options["WISDEM"]["FixedBottomSE"]["n_height"] = len(
                self.wt_init["components"]["monopile"]["outer_shape_bem"]["outer_diameter"]["grid"]
            )
            self.modeling_options["WISDEM"]["FixedBottomSE"]["n_layers"] = len(
                self.wt_init["components"]["monopile"]["internal_structure_2d_fem"]["layers"]
            )
            self.modeling_options["WISDEM"]["FixedBottomSE"]["n_height_monopile"] = self.modeling_options["WISDEM"][
                "FixedBottomSE"
            ]["n_height"]
            self.modeling_options["WISDEM"]["FixedBottomSE"]["n_layers_monopile"] = self.modeling_options["WISDEM"][
                "FixedBottomSE"
            ]["n_layers"]

        # Jacket
        if self.modeling_options["flags"]["jacket"]:
            self.modeling_options["WISDEM"]["FixedBottomSE"]["n_legs"] = self.wt_init["components"]["jacket"]["n_legs"]
            self.modeling_options["WISDEM"]["FixedBottomSE"]["n_bays"] = self.wt_init["components"]["jacket"]["n_bays"]
            self.modeling_options["WISDEM"]["FixedBottomSE"]["mud_brace"] = self.wt_init["components"]["jacket"]["x_mb"]
            self.modeling_options["WISDEM"]["FixedBottomSE"]["material"] = self.wt_init["components"]["jacket"][
                "material"
            ]

        # Floating platform
        self.modeling_options["floating"] = {}
        if self.modeling_options["flags"]["floating_platform"]:
            n_joints = len(self.wt_init["components"]["floating_platform"]["joints"])
            self.modeling_options["floating"]["joints"] = {}
            self.modeling_options["floating"]["joints"]["n_joints"] = n_joints
            self.modeling_options["floating"]["joints"]["name"] = [""] * n_joints
            self.modeling_options["floating"]["joints"]["transition"] = [False] * n_joints
            self.modeling_options["floating"]["joints"]["cylindrical"] = [False] * n_joints
            for i in range(n_joints):
                self.modeling_options["floating"]["joints"]["name"][i] = self.wt_init["components"][
                    "floating_platform"
                ]["joints"][i]["name"]
                self.modeling_options["floating"]["joints"]["transition"][i] = self.wt_init["components"][
                    "floating_platform"
                ]["joints"][i]["transition"]
                self.modeling_options["floating"]["joints"]["cylindrical"][i] = self.wt_init["components"][
                    "floating_platform"
                ]["joints"][i]["cylindrical"]

            # Create name->index dictionary for joint names, will add on axial joints later
            name2idx = dict(zip(self.modeling_options["floating"]["joints"]["name"], range(n_joints)))

            # Check that there is at most one transition joint
            if self.modeling_options["floating"]["joints"]["transition"].count(True) > 1:
                raise ValueError("Can only support one tower on the floating platform for now")
            try:
                itrans = self.modeling_options["floating"]["joints"]["transition"].index(True)
                self.modeling_options["floating"]["transition_joint"] = itrans
            except:
                self.modeling_options["floating"]["transition_joint"] = None

            n_members = len(self.wt_init["components"]["floating_platform"]["members"])
            self.modeling_options["floating"]["members"] = {}
            self.modeling_options["floating"]["members"]["n_members"] = n_members
            self.modeling_options["floating"]["members"]["name"] = [""] * n_members
            self.modeling_options["floating"]["members"]["joint1"] = [""] * n_members
            self.modeling_options["floating"]["members"]["joint2"] = [""] * n_members
            self.modeling_options["floating"]["members"]["outer_shape"] = [""] * n_members
            self.modeling_options["floating"]["members"]["n_height"] = np.zeros(n_members, dtype=int)
            self.modeling_options["floating"]["members"]["n_geom"] = np.zeros(n_members, dtype=int)
            self.modeling_options["floating"]["members"]["n_layers"] = np.zeros(n_members, dtype=int)
            self.modeling_options["floating"]["members"]["n_ballasts"] = np.zeros(n_members, dtype=int)
            self.modeling_options["floating"]["members"]["n_bulkheads"] = np.zeros(n_members, dtype=int)
            self.modeling_options["floating"]["members"]["n_axial_joints"] = np.zeros(n_members, dtype=int)
            ballast_types = []
            for i in range(n_members):
                self.modeling_options["floating"]["members"]["name"][i] = self.wt_init["components"][
                    "floating_platform"
                ]["members"][i]["name"]
                self.modeling_options["floating"]["members"]["joint1"][i] = self.wt_init["components"][
                    "floating_platform"
                ]["members"][i]["joint1"]
                self.modeling_options["floating"]["members"]["joint2"][i] = self.wt_init["components"][
                    "floating_platform"
                ]["members"][i]["joint2"]
                self.modeling_options["floating"]["members"]["outer_shape"][i] = self.wt_init["components"][
                    "floating_platform"
                ]["members"][i]["outer_shape"]["shape"]

                # Master grid for all bulkheads, internal joints, ballasts, geometry changes, etc
                grid = self.wt_init["components"]["floating_platform"]["members"][i]["outer_shape"]["outer_diameter"][
                    "grid"
                ][:]

                # Grid for just diameter / thickness design
                geom_grid = grid[:]

                n_layers = len(
                    self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]["layers"]
                )
                self.modeling_options["floating"]["members"]["n_layers"][i] = n_layers
                if "ballasts" in self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]:
                    n_ballasts = len(
                        self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]["ballasts"]
                    )
                else:
                    n_ballasts = 0
                self.modeling_options["floating"]["members"]["n_ballasts"][i] = n_ballasts

                # Add in bulkheads and enforce at least endcaps for submerged environment
                # Don't add to master grid as they are handled differently in FloatingSE
                if "bulkhead" in self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]:
                    bulkgrid = self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"][
                        "bulkhead"
                    ]["thickness"]["grid"]
                    if not 0.0 in bulkgrid:
                        self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]["bulkhead"][
                            "thickness"
                        ]["grid"].append(0.0)
                        self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]["bulkhead"][
                            "thickness"
                        ]["values"].append(0.02)
                    if not 1.0 in bulkgrid:
                        self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]["bulkhead"][
                            "thickness"
                        ]["grid"].append(1.0)
                        self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]["bulkhead"][
                            "thickness"
                        ]["values"].append(0.02)
                    # grid += bulkgrid # Handled differently in the floating code
                else:
                    self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]["bulkhead"] = {}
                    self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]["bulkhead"][
                        "material"
                    ] = self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]["layers"][
                        0
                    ][
                        "material"
                    ]
                    self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]["bulkhead"][
                        "thickness"
                    ] = {}
                    self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]["bulkhead"][
                        "thickness"
                    ]["grid"] = [0.0, 1.0]
                    self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]["bulkhead"][
                        "thickness"
                    ]["values"] = [0.02, 0.02]

                n_bulk = len(
                    self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"]["bulkhead"][
                        "thickness"
                    ]["grid"]
                )
                self.modeling_options["floating"]["members"]["n_bulkheads"][i] = n_bulk

                self.modeling_options["floating"]["members"][
                    "layer_mat_member_" + self.modeling_options["floating"]["members"]["name"][i]
                ] = [""] * n_layers
                for j in range(n_layers):
                    self.modeling_options["floating"]["members"][
                        "layer_mat_member_" + self.modeling_options["floating"]["members"]["name"][i]
                    ][j] = self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"][
                        "layers"
                    ][
                        j
                    ][
                        "material"
                    ]
                    grid += self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"][
                        "layers"
                    ][j]["thickness"]["grid"]
                    geom_grid += self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"][
                        "layers"
                    ][j]["thickness"]["grid"]

                self.modeling_options["floating"]["members"][
                    "ballast_flag_member_" + self.modeling_options["floating"]["members"]["name"][i]
                ] = [False] * n_ballasts
                self.modeling_options["floating"]["members"][
                    "ballast_mat_member_" + self.modeling_options["floating"]["members"]["name"][i]
                ] = [""] * n_ballasts
                for k in range(n_ballasts):
                    self.modeling_options["floating"]["members"][
                        "ballast_flag_member_" + self.modeling_options["floating"]["members"]["name"][i]
                    ][k] = self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"][
                        "ballasts"
                    ][
                        k
                    ][
                        "variable_flag"
                    ]
                    if (
                        self.modeling_options["floating"]["members"][
                            "ballast_flag_member_" + self.modeling_options["floating"]["members"]["name"][i]
                        ][k]
                        == False
                    ):
                        ballast_types.append(
                            self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"][
                                "ballasts"
                            ][k]["material"]
                        )
                        self.modeling_options["floating"]["members"][
                            "ballast_mat_member_" + self.modeling_options["floating"]["members"]["name"][i]
                        ][k] = ballast_types[-1]
                    else:
                        ballast_types.append("variable")

                    grid += self.wt_init["components"]["floating_platform"]["members"][i]["internal_structure"][
                        "ballasts"
                    ][k]["grid"]

                if "axial_joints" in self.wt_init["components"]["floating_platform"]["members"][i]:
                    n_axial_joints = len(self.wt_init["components"]["floating_platform"]["members"][i]["axial_joints"])
                    self.modeling_options["floating"]["members"]["n_axial_joints"][i] = n_axial_joints
                    self.modeling_options["floating"]["members"][
                        "axial_joint_name_member_" + self.modeling_options["floating"]["members"]["name"][i]
                    ] = [""] * n_axial_joints
                    for m in range(n_axial_joints):
                        self.modeling_options["floating"]["members"][
                            "axial_joint_name_member_" + self.modeling_options["floating"]["members"]["name"][i]
                        ][m] = self.wt_init["components"]["floating_platform"]["members"][i]["axial_joints"][m]["name"]
                        grid.append(
                            self.wt_init["components"]["floating_platform"]["members"][i]["axial_joints"][m]["grid"]
                        )
                        name2idx[
                            self.wt_init["components"]["floating_platform"]["members"][i]["axial_joints"][m]["name"]
                        ] = len(name2idx)
                else:
                    self.modeling_options["floating"]["members"]["n_axial_joints"][i] = 0

                final_grid = np.unique(grid)
                final_geom_grid = np.unique(geom_grid)
                self.modeling_options["floating"]["members"][
                    "grid_member_" + self.modeling_options["floating"]["members"]["name"][i]
                ] = final_grid
                self.modeling_options["floating"]["members"][
                    "geom_member_" + self.modeling_options["floating"]["members"]["name"][i]
                ] = final_geom_grid
                self.modeling_options["floating"]["members"]["n_height"][i] = len(final_grid)
                self.modeling_options["floating"]["members"]["n_geom"][i] = len(final_geom_grid)

            self.modeling_options["floating"]["members"]["ballast_types"] = set(ballast_types)

            # Store joint info
            self.modeling_options["floating"]["joints"]["name2idx"] = name2idx

            # Floating tower params
            self.modeling_options["floating"]["tower"] = {}
            self.modeling_options["floating"]["tower"]["n_ballasts"] = [0]
            self.modeling_options["floating"]["tower"]["n_bulkheads"] = [0]
            self.modeling_options["floating"]["tower"]["n_axial_joints"] = [0]
            if self.modeling_options["flags"]["tower"]:
                self.modeling_options["floating"]["tower"]["n_height"] = [
                    self.modeling_options["WISDEM"]["TowerSE"]["n_height_tower"]
                ]
                self.modeling_options["floating"]["tower"]["n_layers"] = [
                    self.modeling_options["WISDEM"]["TowerSE"]["n_layers_tower"]
                ]
            else:
                self.modeling_options["floating"]["tower"]["n_height"] = [0]
                self.modeling_options["floating"]["tower"]["n_layers"] = [0]

        # Mooring
        self.modeling_options["mooring"] = {}
        if self.modeling_options["flags"]["mooring"]:
            n_nodes = len(self.wt_init["components"]["mooring"]["nodes"])
            n_lines = len(self.wt_init["components"]["mooring"]["lines"])
            n_line_types = len(self.wt_init["components"]["mooring"]["line_types"])
            n_anchor_types = len(self.wt_init["components"]["mooring"]["anchor_types"])
            self.modeling_options["mooring"]["symmetric"] = self.modeling_options["WISDEM"]["FloatingSE"][
                "symmetric_moorings"
            ]
            self.modeling_options["mooring"]["n_nodes"] = n_nodes
            self.modeling_options["mooring"]["n_lines"] = n_lines
            self.modeling_options["mooring"]["n_anchors"] = n_lines
            self.modeling_options["mooring"]["n_line_types"] = n_line_types
            self.modeling_options["mooring"]["n_anchor_types"] = n_anchor_types
            self.modeling_options["mooring"]["node_type"] = [""] * n_nodes
            self.modeling_options["mooring"]["node_names"] = [""] * n_nodes
            self.modeling_options["mooring"]["anchor_type"] = [""] * n_nodes
            self.modeling_options["mooring"]["fairlead_type"] = [""] * n_nodes
            for i in range(n_nodes):
                self.modeling_options["mooring"]["node_type"][i] = self.wt_init["components"]["mooring"]["nodes"][i][
                    "node_type"
                ]
                self.modeling_options["mooring"]["node_names"][i] = self.wt_init["components"]["mooring"]["nodes"][i][
                    "name"
                ]
                self.modeling_options["mooring"]["anchor_type"][i] = self.wt_init["components"]["mooring"]["nodes"][i][
                    "anchor_type"
                ]
                self.modeling_options["mooring"]["fairlead_type"][i] = self.wt_init["components"]["mooring"]["nodes"][
                    i
                ]["fairlead_type"]
            self.modeling_options["mooring"]["node1"] = [""] * n_lines
            self.modeling_options["mooring"]["node2"] = [""] * n_lines
            self.modeling_options["mooring"]["line_type"] = [""] * n_lines
            self.modeling_options["mooring"]["line_material"] = [""] * n_lines
            self.modeling_options["mooring"]["line_anchor"] = [""] * n_lines
            fairlead_nodes = []
            for i in range(n_lines):
                self.modeling_options["mooring"]["node1"][i] = self.wt_init["components"]["mooring"]["lines"][i][
                    "node1"
                ]
                self.modeling_options["mooring"]["node2"][i] = self.wt_init["components"]["mooring"]["lines"][i][
                    "node2"
                ]
                self.modeling_options["mooring"]["line_type"][i] = self.wt_init["components"]["mooring"]["lines"][i][
                    "line_type"
                ]
                # For the vessel attachments, find the list of fairlead nodes on the structure
                node1id = self.modeling_options["mooring"]["node_names"].index(
                    self.modeling_options["mooring"]["node1"][i]
                )
                node2id = self.modeling_options["mooring"]["node_names"].index(
                    self.modeling_options["mooring"]["node2"][i]
                )
                if self.modeling_options["mooring"]["node_type"][node1id] == "vessel":
                    fairlead_nodes.append(self.wt_init["components"]["mooring"]["nodes"][node1id]["joint"])
                if self.modeling_options["mooring"]["node_type"][node2id] == "vessel":
                    fairlead_nodes.append(self.wt_init["components"]["mooring"]["nodes"][node2id]["joint"])
                # Store the anchor type names to start
                if "fix" in self.modeling_options["mooring"]["node_type"][node1id]:
                    self.modeling_options["mooring"]["line_anchor"][i] = self.modeling_options["mooring"][
                        "anchor_type"
                    ][node1id]
                if "fix" in self.modeling_options["mooring"]["node_type"][node2id]:
                    self.modeling_options["mooring"]["line_anchor"][i] = self.modeling_options["mooring"][
                        "anchor_type"
                    ][node2id]

            self.modeling_options["mooring"]["line_type_name"] = [""] * n_line_types
            self.modeling_options["mooring"]["line_type_type"] = [""] * n_line_types
            for i in range(n_line_types):
                self.modeling_options["mooring"]["line_type_name"][i] = self.wt_init["components"]["mooring"][
                    "line_types"
                ][i]["name"]
                self.modeling_options["mooring"]["line_type_type"][i] = self.wt_init["components"]["mooring"][
                    "line_types"
                ][i]["type"].lower()
                for j in range(n_lines):
                    if (
                        self.modeling_options["mooring"]["line_type"][j]
                        == self.modeling_options["mooring"]["line_type_name"][i]
                    ):
                        self.modeling_options["mooring"]["line_material"][j] = self.modeling_options["mooring"][
                            "line_type_type"
                        ][i]
            self.modeling_options["mooring"]["anchor_type_name"] = [""] * n_anchor_types
            self.modeling_options["mooring"]["anchor_type_type"] = [""] * n_anchor_types
            for i in range(n_anchor_types):
                self.modeling_options["mooring"]["anchor_type_name"][i] = self.wt_init["components"]["mooring"][
                    "anchor_types"
                ][i]["name"]
                self.modeling_options["mooring"]["anchor_type_type"][i] = self.wt_init["components"]["mooring"][
                    "anchor_types"
                ][i]["type"].lower()
                for j in range(n_lines):
                    if (
                        self.modeling_options["mooring"]["line_anchor"][j]
                        == self.modeling_options["mooring"]["anchor_type_name"][i]
                    ):
                        self.modeling_options["mooring"]["line_anchor"][j] = self.modeling_options["mooring"][
                            "anchor_type_type"
                        ][i]
            self.modeling_options["mooring"]["n_attach"] = len(set(fairlead_nodes))

        # Assembly
        self.modeling_options["assembly"] = {}
        self.modeling_options["assembly"]["number_of_blades"] = int(self.wt_init["assembly"]["number_of_blades"])

    def set_opt_flags(self):
        # Recursively look for flags to set global optimization flag
        def recursive_flag(d):
            opt_flag = False
            for k, v in d.items():
                if isinstance(v, dict):
                    opt_flag = opt_flag or recursive_flag(v)
                elif k == "flag":
                    opt_flag = opt_flag or v
            return opt_flag

        # The user can provide `opt_flag` in analysis_options.yaml,
        # but if it's not provided, we check the individual opt flags
        # from analysis_options.yaml and set a global `opt_flag`
        if "opt_flag" in self.analysis_options["driver"]:
            self.analysis_options["opt_flag"] = self.analysis_options["driver"]["opt_flag"]
        else:
            self.analysis_options["opt_flag"] = recursive_flag(self.analysis_options["design_variables"])

        # If not an optimization DV, then the number of points should be same as the discretization
        blade_opt_options = self.analysis_options["design_variables"]["blade"]
        if (
            not blade_opt_options["aero_shape"]["twist"]["flag"]
            and not blade_opt_options["aero_shape"]["twist"]["inverse"]
        ):
            blade_opt_options["aero_shape"]["twist"]["n_opt"] = self.modeling_options["WISDEM"]["RotorSE"]["n_span"]
        elif blade_opt_options["aero_shape"]["twist"]["n_opt"] < 4:
            raise ValueError("Cannot optimize twist with less than 4 control points along blade span")

        if not blade_opt_options["aero_shape"]["chord"]["flag"]:
            blade_opt_options["aero_shape"]["chord"]["n_opt"] = self.modeling_options["WISDEM"]["RotorSE"]["n_span"]
        elif blade_opt_options["aero_shape"]["chord"]["n_opt"] < 4:
            raise ValueError("Cannot optimize chord with less than 4 control points along blade span")

        if not blade_opt_options["aero_shape"]["t/c"]["flag"]:
            blade_opt_options["aero_shape"]["t/c"]["n_opt"] = self.modeling_options["WISDEM"]["RotorSE"]["n_span"]
        elif blade_opt_options["aero_shape"]["t/c"]["n_opt"] < 4:
            raise ValueError("Cannot optimize t/c with less than 4 control points along blade span")

        if not blade_opt_options["aero_shape"]["L/D"]["flag"]:
            blade_opt_options["aero_shape"]["L/D"]["n_opt"] = self.modeling_options["WISDEM"]["RotorSE"]["n_span"]
        elif blade_opt_options["aero_shape"]["L/D"]["n_opt"] < 4:
            raise ValueError("Cannot optimize L/D with less than 4 control points along blade span")

        if not blade_opt_options["structure"]["spar_cap_ss"]["flag"]:
            blade_opt_options["structure"]["spar_cap_ss"]["n_opt"] = self.modeling_options["WISDEM"]["RotorSE"][
                "n_span"
            ]
        elif blade_opt_options["structure"]["spar_cap_ss"]["n_opt"] < 4:
            raise ValueError("Cannot optimize spar cap suction side with less than 4 control points along blade span")

        if not blade_opt_options["structure"]["spar_cap_ps"]["flag"]:
            blade_opt_options["structure"]["spar_cap_ps"]["n_opt"] = self.modeling_options["WISDEM"]["RotorSE"][
                "n_span"
            ]
        elif blade_opt_options["structure"]["spar_cap_ps"]["n_opt"] < 4:
            raise ValueError("Cannot optimize spar cap pressure side with less than 4 control points along blade span")

        if not blade_opt_options["structure"]["te_ss"]["flag"]:
            blade_opt_options["structure"]["te_ss"]["n_opt"] = self.modeling_options["WISDEM"]["RotorSE"]["n_span"]
        elif blade_opt_options["structure"]["te_ss"]["n_opt"] < 4:
            raise ValueError(
                "Cannot optimize trailing edge suction side with less than 4 control points along blade span"
            )

        if not blade_opt_options["structure"]["te_ps"]["flag"]:
            blade_opt_options["structure"]["te_ps"]["n_opt"] = self.modeling_options["WISDEM"]["RotorSE"]["n_span"]
        elif blade_opt_options["structure"]["te_ps"]["n_opt"] < 4:
            raise ValueError(
                "Cannot optimize trailing edge pressure side with less than 4 control points along blade span"
            )

        # Handle linked joints and members in floating platform
        if self.modeling_options["flags"]["floating"]:
            float_opt_options = self.analysis_options["design_variables"]["floating"]

            # First the joints
            dv_info = []
            for c in ["z", "r"]:
                for idv in float_opt_options["joints"][c + "_coordinate"]:
                    inames = idv["names"]
                    idx = [self.modeling_options["floating"]["joints"]["name"].index(m) for m in inames]

                    idict = {}
                    idict["indices"] = idx
                    idict["dimension"] = 0 if c == "r" else 2
                    dv_info.append(idict)

            # Check for r-coordinate dv and cylindrical consistency
            for idict in dv_info:
                if idict["dimension"] != 0:
                    continue
                for k in idict["indices"]:
                    m = self.modeling_options["floating"]["joints"]["name"][k]
                    if not self.modeling_options["floating"]["joints"]["cylindrical"][k]:
                        raise ValueError(f"Cannot optimize r-coordinate of, {m}, becase it is not a cylindrical joint")

            # Store DV information for needed linking and IVC assignment
            self.modeling_options["floating"]["joints"]["design_variable_data"] = dv_info

            # Now the members
            memgrps = [[m] for m in self.modeling_options["floating"]["members"]["name"]]
            for idv in float_opt_options["members"]["groups"]:
                inames = idv["names"]
                idx = [self.modeling_options["floating"]["members"]["name"].index(m) for m in inames]
                for k in range(1, len(idx)):
                    try:
                        memgrps[idx[k]].remove(inames[k])
                        memgrps[idx[0]].append(inames[k])
                    except ValueError:
                        raise ValueError("Cannot put member," + inames[k] + ", as part of multiple groups")

            # Remove entries for members that are now linked with others
            while [] in memgrps:
                memgrps.remove([])
            self.modeling_options["floating"]["members"]["linked_members"] = memgrps

            # Make a name 2 group index lookup
            name2grp = {}
            for k, kgrp in enumerate(memgrps):
                for kname in kgrp:
                    name2grp[kname] = k
            self.modeling_options["floating"]["members"]["name2idx"] = name2grp

    def write_ontology(self, wt_opt, fname_output):

        # Update blade
        if self.modeling_options["flags"]["blade"]:
            # Update blade outer shape
            self.wt_init["components"]["blade"]["outer_shape_bem"]["airfoil_position"]["grid"] = wt_opt[
                "blade.opt_var.af_position"
            ].tolist()
            self.wt_init["components"]["blade"]["outer_shape_bem"]["chord"]["grid"] = wt_opt[
                "blade.outer_shape_bem.s"
            ].tolist()
            self.wt_init["components"]["blade"]["outer_shape_bem"]["chord"]["values"] = wt_opt[
                "blade.pa.chord_param"
            ].tolist()
            self.wt_init["components"]["blade"]["outer_shape_bem"]["twist"]["grid"] = wt_opt[
                "blade.outer_shape_bem.s"
            ].tolist()
            self.wt_init["components"]["blade"]["outer_shape_bem"]["twist"]["values"] = wt_opt["rotorse.theta"].tolist()
            self.wt_init["components"]["blade"]["outer_shape_bem"]["pitch_axis"]["grid"] = wt_opt[
                "blade.outer_shape_bem.s"
            ].tolist()
            self.wt_init["components"]["blade"]["outer_shape_bem"]["pitch_axis"]["values"] = wt_opt[
                "blade.outer_shape_bem.pitch_axis"
            ].tolist()
            if self.modeling_options["WISDEM"]["RotorSE"]["inn_af"]:
                self.wt_init["components"]["blade"]["outer_shape_bem"]["t/c"]["grid"] = wt_opt[
                    "blade.outer_shape_bem.s"
                ].tolist()
                self.wt_init["components"]["blade"]["outer_shape_bem"]["t/c"]["values"] = wt_opt[
                    "blade.interp_airfoils.r_thick_interp"
                ].tolist()
                self.wt_init["components"]["blade"]["outer_shape_bem"]["L/D"]["grid"] = wt_opt[
                    "blade.outer_shape_bem.s"
                ].tolist()
                self.wt_init["components"]["blade"]["outer_shape_bem"]["L/D"]["values"] = wt_opt[
                    "rotorse.rp.powercurve.L_D"
                ].tolist()
                self.wt_init["components"]["blade"]["outer_shape_bem"]["c_d"]["grid"] = wt_opt[
                    "blade.outer_shape_bem.s"
                ].tolist()
                self.wt_init["components"]["blade"]["outer_shape_bem"]["c_d"]["values"] = wt_opt[
                    "rotorse.rp.powercurve.cd_regII"
                ].tolist()
                self.wt_init["components"]["blade"]["outer_shape_bem"]["stall_margin"]["grid"] = wt_opt[
                    "blade.outer_shape_bem.s"
                ].tolist()
                stall_margin = np.deg2rad(
                    wt_opt["rotorse.stall_check.stall_angle_along_span"] - wt_opt["rotorse.stall_check.aoa_along_span"]
                )
                self.wt_init["components"]["blade"]["outer_shape_bem"]["stall_margin"]["values"] = stall_margin.tolist()
            self.wt_init["components"]["blade"]["outer_shape_bem"]["reference_axis"]["x"]["grid"] = wt_opt[
                "blade.outer_shape_bem.s"
            ].tolist()
            self.wt_init["components"]["blade"]["outer_shape_bem"]["reference_axis"]["y"]["grid"] = wt_opt[
                "blade.outer_shape_bem.s"
            ].tolist()
            self.wt_init["components"]["blade"]["outer_shape_bem"]["reference_axis"]["z"]["grid"] = wt_opt[
                "blade.outer_shape_bem.s"
            ].tolist()
            self.wt_init["components"]["blade"]["outer_shape_bem"]["reference_axis"]["x"]["values"] = wt_opt[
                "blade.outer_shape_bem.ref_axis"
            ][:, 0].tolist()
            self.wt_init["components"]["blade"]["outer_shape_bem"]["reference_axis"]["y"]["values"] = wt_opt[
                "blade.outer_shape_bem.ref_axis"
            ][:, 1].tolist()
            self.wt_init["components"]["blade"]["outer_shape_bem"]["reference_axis"]["z"]["values"] = wt_opt[
                "blade.outer_shape_bem.ref_axis"
            ][:, 2].tolist()

            # Update blade structure
            # Reference axis from blade outer shape
            self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["reference_axis"] = self.wt_init[
                "components"
            ]["blade"]["outer_shape_bem"]["reference_axis"]
            # Webs positions
            for i in range(self.modeling_options["WISDEM"]["RotorSE"]["n_webs"]):
                if "rotation" in self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["webs"]:
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["webs"][i]["rotation"][
                        "grid"
                    ] = wt_opt["blade.internal_structure_2d_fem.s"].tolist()
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["webs"][i]["rotation"][
                        "values"
                    ] = wt_opt["blade.internal_structure_2d_fem.web_rotation"][i, :].tolist()
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["webs"][i]["offset_y_pa"][
                        "grid"
                    ] = wt_opt["blade.internal_structure_2d_fem.s"].tolist()
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["webs"][i]["offset_y_pa"][
                        "values"
                    ] = wt_opt["blade.internal_structure_2d_fem.web_offset_y_pa"][i, :].tolist()
                if "start_nd_arc" not in self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["webs"][i]:
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["webs"][i]["start_nd_arc"] = {}
                if "end_nd_arc" not in self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["webs"][i]:
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["webs"][i]["end_nd_arc"] = {}
                self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["webs"][i]["start_nd_arc"][
                    "grid"
                ] = wt_opt["blade.internal_structure_2d_fem.s"].tolist()
                self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["webs"][i]["start_nd_arc"][
                    "values"
                ] = wt_opt["blade.internal_structure_2d_fem.web_start_nd"][i, :].tolist()
                self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["webs"][i]["end_nd_arc"][
                    "grid"
                ] = wt_opt["blade.internal_structure_2d_fem.s"].tolist()
                self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["webs"][i]["end_nd_arc"][
                    "values"
                ] = wt_opt["blade.internal_structure_2d_fem.web_end_nd"][i, :].tolist()

            # Structural layers
            for i in range(self.modeling_options["WISDEM"]["RotorSE"]["n_layers"]):
                self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["thickness"][
                    "grid"
                ] = wt_opt["blade.internal_structure_2d_fem.s"].tolist()
                self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["thickness"][
                    "values"
                ] = wt_opt["blade.ps.layer_thickness_param"][i, :].tolist()
                if wt_opt["blade.internal_structure_2d_fem.definition_layer"][i] < 7:
                    if (
                        "start_nd_arc"
                        not in self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]
                    ):
                        self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i][
                            "start_nd_arc"
                        ] = {}
                    if (
                        "end_nd_arc"
                        not in self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]
                    ):
                        self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["end_nd_arc"] = {}
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["start_nd_arc"][
                        "grid"
                    ] = wt_opt["blade.internal_structure_2d_fem.s"].tolist()
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["start_nd_arc"][
                        "values"
                    ] = wt_opt["blade.internal_structure_2d_fem.layer_start_nd"][i, :].tolist()
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["end_nd_arc"][
                        "grid"
                    ] = wt_opt["blade.internal_structure_2d_fem.s"].tolist()
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["end_nd_arc"][
                        "values"
                    ] = wt_opt["blade.internal_structure_2d_fem.layer_end_nd"][i, :].tolist()
                if (
                    wt_opt["blade.internal_structure_2d_fem.definition_layer"][i] > 1
                    and wt_opt["blade.internal_structure_2d_fem.definition_layer"][i] < 6
                ):
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["width"][
                        "grid"
                    ] = wt_opt["blade.internal_structure_2d_fem.s"].tolist()
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["width"][
                        "values"
                    ] = wt_opt["blade.internal_structure_2d_fem.layer_width"][i, :].tolist()
                if (
                    wt_opt["blade.internal_structure_2d_fem.definition_layer"][i] == 2
                    or wt_opt["blade.internal_structure_2d_fem.definition_layer"][i] == 3
                ):
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["rotation"][
                        "grid"
                    ] = wt_opt["blade.internal_structure_2d_fem.s"].tolist()
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["rotation"][
                        "values"
                    ] = wt_opt["blade.internal_structure_2d_fem.layer_rotation"][i, :].tolist()
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["offset_y_pa"][
                        "grid"
                    ] = wt_opt["blade.internal_structure_2d_fem.s"].tolist()
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["offset_y_pa"][
                        "values"
                    ] = wt_opt["blade.internal_structure_2d_fem.layer_offset_y_pa"][i, :].tolist()
                if (
                    wt_opt["blade.internal_structure_2d_fem.definition_layer"][i] == 4
                    or wt_opt["blade.internal_structure_2d_fem.definition_layer"][i] == 5
                ):
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["midpoint_nd_arc"][
                        "grid"
                    ] = wt_opt["blade.internal_structure_2d_fem.s"].tolist()
                    self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["midpoint_nd_arc"][
                        "values"
                    ] = wt_opt["blade.internal_structure_2d_fem.layer_midpoint_nd"][i, :].tolist()

                self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["fiber_orientation"] = {}

                self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["fiber_orientation"][
                    "grid"
                ] = wt_opt["blade.internal_structure_2d_fem.s"].tolist()
                self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["layers"][i]["fiber_orientation"][
                    "values"
                ] = np.zeros(len(wt_opt["blade.internal_structure_2d_fem.s"])).tolist()

            # TODO assign joint mass to wt_init from rs.bjs
            # Elastic properties of the blade
            if self.modeling_options["WISDEM"]["RotorSE"]["bjs"]:
                self.wt_init["components"]["blade"]["internal_structure_2d_fem"]["joint"]["mass"] = float(
                    wt_opt["rotorse.rs.bjs.joint_mass"]
                )

            self.wt_init["components"]["blade"]["elastic_properties_mb"] = {}
            self.wt_init["components"]["blade"]["elastic_properties_mb"]["six_x_six"] = {}
            self.wt_init["components"]["blade"]["elastic_properties_mb"]["six_x_six"]["reference_axis"] = self.wt_init[
                "components"
            ]["blade"]["internal_structure_2d_fem"]["reference_axis"]
            self.wt_init["components"]["blade"]["elastic_properties_mb"]["six_x_six"]["twist"] = self.wt_init[
                "components"
            ]["blade"]["outer_shape_bem"]["twist"]
            self.wt_init["components"]["blade"]["elastic_properties_mb"]["six_x_six"]["stiff_matrix"] = {}
            self.wt_init["components"]["blade"]["elastic_properties_mb"]["six_x_six"]["stiff_matrix"]["grid"] = wt_opt[
                "blade.outer_shape_bem.s"
            ].tolist()
            K = []
            for i in range(self.modeling_options["WISDEM"]["RotorSE"]["n_span"]):
                Ki = np.zeros(21)
                Ki[11] = wt_opt["rotorse.EA"][i]
                Ki[15] = wt_opt["rotorse.EIxx"][i]
                Ki[18] = wt_opt["rotorse.EIyy"][i]
                Ki[20] = wt_opt["rotorse.GJ"][i]
                K.append(Ki.tolist())
            self.wt_init["components"]["blade"]["elastic_properties_mb"]["six_x_six"]["stiff_matrix"]["values"] = K
            self.wt_init["components"]["blade"]["elastic_properties_mb"]["six_x_six"]["inertia_matrix"] = {}
            self.wt_init["components"]["blade"]["elastic_properties_mb"]["six_x_six"]["inertia_matrix"][
                "grid"
            ] = wt_opt["blade.outer_shape_bem.s"].tolist()
            I = []
            for i in range(self.modeling_options["WISDEM"]["RotorSE"]["n_span"]):
                Ii = np.zeros(21)
                Ii[0] = wt_opt["rotorse.rhoA"][i]
                Ii[5] = -wt_opt["rotorse.rhoA"][i] * wt_opt["rotorse.re.y_cg"][i]
                Ii[6] = wt_opt["rotorse.rhoA"][i]
                Ii[10] = wt_opt["rotorse.rhoA"][i] * wt_opt["rotorse.re.x_cg"][i]
                Ii[11] = wt_opt["rotorse.rhoA"][i]
                Ii[12] = wt_opt["rotorse.rhoA"][i] * wt_opt["rotorse.re.y_cg"][i]
                Ii[13] = -wt_opt["rotorse.rhoA"][i] * wt_opt["rotorse.re.x_cg"][i]
                Ii[15] = wt_opt["rotorse.re.precomp.edge_iner"][i]
                #Ii[16] = wt_opt["rotorse.re.precomp.edge_iner"][i]
                Ii[18] = wt_opt['rotorse.re.precomp.flap_iner'][i]
                Ii[20] = wt_opt["rotorse.rhoJ"][i]
                I.append(Ii.tolist())
            self.wt_init["components"]["blade"]["elastic_properties_mb"]["six_x_six"]["inertia_matrix"]["values"] = I

        # Update hub
        if self.modeling_options["flags"]["hub"]:
            # Update hub
            self.wt_init["components"]["hub"]["cone_angle"] = float(wt_opt["hub.cone"])
            self.wt_init["components"]["hub"]["flange_t2shell_t"] = float(wt_opt["hub.flange_t2shell_t"])
            self.wt_init["components"]["hub"]["flange_OD2hub_D"] = float(wt_opt["hub.flange_OD2hub_D"])
            self.wt_init["components"]["hub"]["flange_ID2OD"] = float(wt_opt["hub.flange_ID2flange_OD"])
            self.wt_init["components"]["hub"]["hub_blade_spacing_margin"] = float(wt_opt["hub.hub_in2out_circ"])
            self.wt_init["components"]["hub"]["hub_stress_concentration"] = float(
                wt_opt["hub.hub_stress_concentration"]
            )
            self.wt_init["components"]["hub"]["n_front_brackets"] = int(wt_opt["hub.n_front_brackets"])
            self.wt_init["components"]["hub"]["n_rear_brackets"] = int(wt_opt["hub.n_rear_brackets"])
            self.wt_init["components"]["hub"]["clearance_hub_spinner"] = float(wt_opt["hub.clearance_hub_spinner"])
            self.wt_init["components"]["hub"]["spin_hole_incr"] = float(wt_opt["hub.spin_hole_incr"])
            self.wt_init["components"]["hub"]["pitch_system_scaling_factor"] = float(
                wt_opt["hub.pitch_system_scaling_factor"]
            )
            self.wt_init["components"]["hub"]["spinner_gust_ws"] = float(wt_opt["hub.spinner_gust_ws"])

        # Update nacelle
        if self.modeling_options["flags"]["nacelle"]:
            # Common direct and geared
            self.wt_init["components"]["nacelle"]["drivetrain"]["uptilt"] = float(wt_opt["nacelle.uptilt"])
            self.wt_init["components"]["nacelle"]["drivetrain"]["distance_tt_hub"] = float(
                wt_opt["nacelle.distance_tt_hub"]
            )
            self.wt_init["components"]["nacelle"]["drivetrain"]["overhang"] = float(wt_opt["nacelle.overhang"])
            self.wt_init["components"]["nacelle"]["drivetrain"]["distance_hub_mb"] = float(
                wt_opt["nacelle.distance_hub2mb"]
            )
            self.wt_init["components"]["nacelle"]["drivetrain"]["distance_mb_mb"] = float(
                wt_opt["nacelle.distance_mb2mb"]
            )
            self.wt_init["components"]["nacelle"]["drivetrain"]["generator_length"] = float(
                wt_opt["nacelle.L_generator"]
            )
            s_lss = np.linspace(0.0, 1.0, len(wt_opt["nacelle.lss_diameter"])).tolist()
            self.wt_init["components"]["nacelle"]["drivetrain"]["lss_diameter"] = wt_opt[
                "nacelle.lss_diameter"
            ].tolist()
            self.wt_init["components"]["nacelle"]["drivetrain"]["lss_wall_thickness"] = wt_opt[
                "nacelle.lss_wall_thickness"
            ].tolist()
            self.wt_init["components"]["nacelle"]["drivetrain"]["gear_ratio"] = float(wt_opt["nacelle.gear_ratio"])
            self.wt_init["components"]["nacelle"]["drivetrain"]["gearbox_efficiency"] = float(
                wt_opt["nacelle.gearbox_efficiency"]
            )
            self.wt_init["components"]["nacelle"]["drivetrain"]["mb1Type"] = wt_opt["nacelle.mb1Type"]
            self.wt_init["components"]["nacelle"]["drivetrain"]["mb2Type"] = wt_opt["nacelle.mb2Type"]
            self.wt_init["components"]["nacelle"]["drivetrain"]["uptower"] = wt_opt["nacelle.uptower"]
            self.wt_init["components"]["nacelle"]["drivetrain"]["lss_material"] = wt_opt["nacelle.lss_material"]
            self.wt_init["components"]["nacelle"]["drivetrain"]["bedplate_material"] = wt_opt[
                "nacelle.bedplate_material"
            ]

            if self.modeling_options["WISDEM"]["DriveSE"]["direct"]:
                # Direct only
                s_nose = np.linspace(0.0, 1.0, len(wt_opt["nacelle.nose_diameter"])).tolist()
                s_bed = np.linspace(0.0, 1.0, len(wt_opt["nacelle.bedplate_wall_thickness"])).tolist()
                self.wt_init["components"]["nacelle"]["drivetrain"]["nose_diameter"] = wt_opt[
                    "nacelle.nose_diameter"
                ].tolist()
                self.wt_init["components"]["nacelle"]["drivetrain"]["nose_wall_thickness"] = wt_opt[
                    "nacelle.nose_wall_thickness"
                ].tolist()
                self.wt_init["components"]["nacelle"]["drivetrain"]["bedplate_wall_thickness"]["grid"] = s_bed
                self.wt_init["components"]["nacelle"]["drivetrain"]["bedplate_wall_thickness"]["values"] = wt_opt[
                    "nacelle.bedplate_wall_thickness"
                ].tolist()
            else:
                # Geared only
                s_hss = np.linspace(0.0, 1.0, len(wt_opt["nacelle.hss_diameter"])).tolist()
                self.wt_init["components"]["nacelle"]["drivetrain"]["hss_length"] = float(wt_opt["nacelle.hss_length"])
                self.wt_init["components"]["nacelle"]["drivetrain"]["hss_diameter"] = wt_opt[
                    "nacelle.hss_diameter"
                ].tolist()
                self.wt_init["components"]["nacelle"]["drivetrain"]["hss_wall_thickness"] = wt_opt[
                    "nacelle.hss_wall_thickness"
                ].tolist()
                self.wt_init["components"]["nacelle"]["drivetrain"]["bedplate_flange_width"] = float(
                    wt_opt["nacelle.bedplate_flange_width"]
                )
                self.wt_init["components"]["nacelle"]["drivetrain"]["bedplate_flange_thickness"] = float(
                    wt_opt["nacelle.bedplate_flange_thickness"]
                )
                self.wt_init["components"]["nacelle"]["drivetrain"]["bedplate_web_thickness"] = float(
                    wt_opt["nacelle.bedplate_web_thickness"]
                )
                self.wt_init["components"]["nacelle"]["drivetrain"]["gear_configuration"] = wt_opt[
                    "nacelle.gear_configuration"
                ]
                self.wt_init["components"]["nacelle"]["drivetrain"]["planet_numbers"] = wt_opt["nacelle.planet_numbers"]
                self.wt_init["components"]["nacelle"]["drivetrain"]["hss_material"] = wt_opt["nacelle.hss_material"]

        # Update generator
        if self.modeling_options["flags"]["generator"]:

            self.wt_init["components"]["nacelle"]["generator"]["B_r"] = float(wt_opt["generator.B_r"])
            self.wt_init["components"]["nacelle"]["generator"]["P_Fe0e"] = float(wt_opt["generator.P_Fe0e"])
            self.wt_init["components"]["nacelle"]["generator"]["P_Fe0h"] = float(wt_opt["generator.P_Fe0h"])
            self.wt_init["components"]["nacelle"]["generator"]["S_N"] = float(wt_opt["generator.S_N"])
            self.wt_init["components"]["nacelle"]["generator"]["alpha_p"] = float(wt_opt["generator.alpha_p"])
            self.wt_init["components"]["nacelle"]["generator"]["b_r_tau_r"] = float(wt_opt["generator.b_r_tau_r"])
            self.wt_init["components"]["nacelle"]["generator"]["b_ro"] = float(wt_opt["generator.b_ro"])
            self.wt_init["components"]["nacelle"]["generator"]["b_s_tau_s"] = float(wt_opt["generator.b_s_tau_s"])
            self.wt_init["components"]["nacelle"]["generator"]["b_so"] = float(wt_opt["generator.b_so"])
            self.wt_init["components"]["nacelle"]["generator"]["cofi"] = float(wt_opt["generator.cofi"])
            self.wt_init["components"]["nacelle"]["generator"]["freq"] = float(wt_opt["generator.freq"])
            self.wt_init["components"]["nacelle"]["generator"]["h_i"] = float(wt_opt["generator.h_i"])
            self.wt_init["components"]["nacelle"]["generator"]["h_sy0"] = float(wt_opt["generator.h_sy0"])
            self.wt_init["components"]["nacelle"]["generator"]["h_w"] = float(wt_opt["generator.h_w"])
            self.wt_init["components"]["nacelle"]["generator"]["k_fes"] = float(wt_opt["generator.k_fes"])
            self.wt_init["components"]["nacelle"]["generator"]["k_fillr"] = float(wt_opt["generator.k_fillr"])
            self.wt_init["components"]["nacelle"]["generator"]["k_fills"] = float(wt_opt["generator.k_fills"])
            self.wt_init["components"]["nacelle"]["generator"]["k_s"] = float(wt_opt["generator.k_s"])
            self.wt_init["components"]["nacelle"]["generator"]["m"] = float(wt_opt["generator.m"])
            self.wt_init["components"]["nacelle"]["generator"]["mu_0"] = float(wt_opt["generator.mu_0"])
            self.wt_init["components"]["nacelle"]["generator"]["mu_r"] = float(wt_opt["generator.mu_r"])
            self.wt_init["components"]["nacelle"]["generator"]["p"] = float(wt_opt["generator.p"])
            self.wt_init["components"]["nacelle"]["generator"]["phi"] = float(wt_opt["generator.phi"])
            self.wt_init["components"]["nacelle"]["generator"]["q1"] = float(wt_opt["generator.q1"])
            self.wt_init["components"]["nacelle"]["generator"]["q2"] = float(wt_opt["generator.q2"])
            self.wt_init["components"]["nacelle"]["generator"]["ratio_mw2pp"] = float(wt_opt["generator.ratio_mw2pp"])
            self.wt_init["components"]["nacelle"]["generator"]["resist_Cu"] = float(wt_opt["generator.resist_Cu"])
            self.wt_init["components"]["nacelle"]["generator"]["sigma"] = float(wt_opt["generator.sigma"])
            self.wt_init["components"]["nacelle"]["generator"]["y_tau_p"] = float(wt_opt["generator.y_tau_p"])
            self.wt_init["components"]["nacelle"]["generator"]["y_tau_pr"] = float(wt_opt["generator.y_tau_pr"])

            self.wt_init["components"]["nacelle"]["generator"]["I_0"] = float(wt_opt["generator.I_0"])
            self.wt_init["components"]["nacelle"]["generator"]["d_r"] = float(wt_opt["generator.d_r"])
            self.wt_init["components"]["nacelle"]["generator"]["h_m"] = float(wt_opt["generator.h_m"])
            self.wt_init["components"]["nacelle"]["generator"]["h_0"] = float(wt_opt["generator.h_0"])
            self.wt_init["components"]["nacelle"]["generator"]["h_s"] = float(wt_opt["generator.h_s"])
            self.wt_init["components"]["nacelle"]["generator"]["len_s"] = float(wt_opt["generator.len_s"])
            self.wt_init["components"]["nacelle"]["generator"]["n_r"] = float(wt_opt["generator.n_r"])
            self.wt_init["components"]["nacelle"]["generator"]["rad_ag"] = float(wt_opt["generator.rad_ag"])
            self.wt_init["components"]["nacelle"]["generator"]["t_wr"] = float(wt_opt["generator.t_wr"])

            self.wt_init["components"]["nacelle"]["generator"]["n_s"] = float(wt_opt["generator.n_s"])
            self.wt_init["components"]["nacelle"]["generator"]["b_st"] = float(wt_opt["generator.b_st"])
            self.wt_init["components"]["nacelle"]["generator"]["d_s"] = float(wt_opt["generator.d_s"])
            self.wt_init["components"]["nacelle"]["generator"]["t_ws"] = float(wt_opt["generator.t_ws"])

            self.wt_init["components"]["nacelle"]["generator"]["rho_Copper"] = float(wt_opt["generator.rho_Copper"])
            self.wt_init["components"]["nacelle"]["generator"]["rho_Fe"] = float(wt_opt["generator.rho_Fe"])
            self.wt_init["components"]["nacelle"]["generator"]["rho_Fes"] = float(wt_opt["generator.rho_Fes"])
            self.wt_init["components"]["nacelle"]["generator"]["rho_PM"] = float(wt_opt["generator.rho_PM"])

            self.wt_init["components"]["nacelle"]["generator"]["C_Cu"] = float(wt_opt["generator.C_Cu"])
            self.wt_init["components"]["nacelle"]["generator"]["C_Fe"] = float(wt_opt["generator.C_Fe"])
            self.wt_init["components"]["nacelle"]["generator"]["C_Fes"] = float(wt_opt["generator.C_Fes"])
            self.wt_init["components"]["nacelle"]["generator"]["C_PM"] = float(wt_opt["generator.C_PM"])

            if self.modeling_options["WISDEM"]["GeneratorSE"]["type"] in ["pmsg_outer"]:
                self.wt_init["components"]["nacelle"]["generator"]["N_c"] = float(wt_opt["generator.N_c"])
                self.wt_init["components"]["nacelle"]["generator"]["b"] = float(wt_opt["generator.b"])
                self.wt_init["components"]["nacelle"]["generator"]["c"] = float(wt_opt["generator.c"])
                self.wt_init["components"]["nacelle"]["generator"]["E_p"] = float(wt_opt["generator.E_p"])
                self.wt_init["components"]["nacelle"]["generator"]["h_yr"] = float(wt_opt["generator.h_yr"])
                self.wt_init["components"]["nacelle"]["generator"]["h_ys"] = float(wt_opt["generator.h_ys"])
                self.wt_init["components"]["nacelle"]["generator"]["h_sr"] = float(wt_opt["generator.h_sr"])
                self.wt_init["components"]["nacelle"]["generator"]["h_ss"] = float(wt_opt["generator.h_ss"])
                self.wt_init["components"]["nacelle"]["generator"]["t_r"] = float(wt_opt["generator.t_r"])
                self.wt_init["components"]["nacelle"]["generator"]["t_s"] = float(wt_opt["generator.t_s"])

                self.wt_init["components"]["nacelle"]["generator"]["u_allow_pcent"] = float(
                    wt_opt["generator.u_allow_pcent"]
                )
                self.wt_init["components"]["nacelle"]["generator"]["y_allow_pcent"] = float(
                    wt_opt["generator.y_allow_pcent"]
                )
                self.wt_init["components"]["nacelle"]["generator"]["z_allow_deg"] = float(
                    wt_opt["generator.z_allow_deg"]
                )
                self.wt_init["components"]["nacelle"]["generator"]["B_tmax"] = float(wt_opt["generator.B_tmax"])

            if self.modeling_options["WISDEM"]["GeneratorSE"]["type"] in ["eesg", "pmsg_arms", "pmsg_disc"]:
                self.wt_init["components"]["nacelle"]["generator"]["tau_p"] = float(wt_opt["generator.tau_p"])
                self.wt_init["components"]["nacelle"]["generator"]["h_ys"] = float(wt_opt["generator.h_ys"])
                self.wt_init["components"]["nacelle"]["generator"]["h_yr"] = float(wt_opt["generator.h_yr"])
                self.wt_init["components"]["nacelle"]["generator"]["b_arm"] = float(wt_opt["generator.b_arm"])

            elif self.modeling_options["WISDEM"]["GeneratorSE"]["type"] in ["scig", "dfig"]:
                self.wt_init["components"]["nacelle"]["generator"]["B_symax"] = float(wt_opt["generator.B_symax"])
                self.wt_init["components"]["nacelle"]["generator"]["S_Nmax"] = float(wt_opt["generator.S_Nmax"])

        # Update tower
        if self.modeling_options["flags"]["tower"]:
            self.wt_init["components"]["tower"]["outer_shape_bem"]["outer_diameter"]["grid"] = wt_opt[
                "tower_grid.s"
            ].tolist()
            self.wt_init["components"]["tower"]["outer_shape_bem"]["outer_diameter"]["values"] = wt_opt[
                "tower.diameter"
            ].tolist()
            self.wt_init["components"]["tower"]["outer_shape_bem"]["reference_axis"]["x"]["grid"] = wt_opt[
                "tower_grid.s"
            ].tolist()
            self.wt_init["components"]["tower"]["outer_shape_bem"]["reference_axis"]["y"]["grid"] = wt_opt[
                "tower_grid.s"
            ].tolist()
            self.wt_init["components"]["tower"]["outer_shape_bem"]["reference_axis"]["z"]["grid"] = wt_opt[
                "tower_grid.s"
            ].tolist()
            self.wt_init["components"]["tower"]["outer_shape_bem"]["reference_axis"]["x"]["values"] = wt_opt[
                "tower.ref_axis"
            ][:, 0].tolist()
            self.wt_init["components"]["tower"]["outer_shape_bem"]["reference_axis"]["y"]["values"] = wt_opt[
                "tower.ref_axis"
            ][:, 1].tolist()
            self.wt_init["components"]["tower"]["outer_shape_bem"]["reference_axis"]["z"]["values"] = wt_opt[
                "tower.ref_axis"
            ][:, 2].tolist()
            self.wt_init["components"]["tower"]["internal_structure_2d_fem"]["outfitting_factor"] = float(
                wt_opt["tower.outfitting_factor"]
            )
            for i in range(self.modeling_options["WISDEM"]["TowerSE"]["n_layers_tower"]):
                self.wt_init["components"]["tower"]["internal_structure_2d_fem"]["layers"][i]["thickness"][
                    "grid"
                ] = wt_opt["tower_grid.s"].tolist()
                self.wt_init["components"]["tower"]["internal_structure_2d_fem"]["layers"][i]["thickness"][
                    "values"
                ] = wt_opt["tower.layer_thickness"][i, :].tolist()

        # Update monopile
        if self.modeling_options["flags"]["monopile"]:
            self.wt_init["components"]["monopile"]["outer_shape_bem"]["outer_diameter"]["grid"] = wt_opt[
                "monopile.s"
            ].tolist()
            self.wt_init["components"]["monopile"]["outer_shape_bem"]["outer_diameter"]["values"] = wt_opt[
                "monopile.diameter"
            ].tolist()
            self.wt_init["components"]["monopile"]["outer_shape_bem"]["reference_axis"]["x"]["grid"] = wt_opt[
                "monopile.s"
            ].tolist()
            self.wt_init["components"]["monopile"]["outer_shape_bem"]["reference_axis"]["y"]["grid"] = wt_opt[
                "monopile.s"
            ].tolist()
            self.wt_init["components"]["monopile"]["outer_shape_bem"]["reference_axis"]["z"]["grid"] = wt_opt[
                "monopile.s"
            ].tolist()
            self.wt_init["components"]["monopile"]["outer_shape_bem"]["reference_axis"]["x"]["values"] = wt_opt[
                "monopile.ref_axis"
            ][:, 0].tolist()
            self.wt_init["components"]["monopile"]["outer_shape_bem"]["reference_axis"]["y"]["values"] = wt_opt[
                "monopile.ref_axis"
            ][:, 1].tolist()
            self.wt_init["components"]["monopile"]["outer_shape_bem"]["reference_axis"]["z"]["values"] = wt_opt[
                "monopile.ref_axis"
            ][:, 2].tolist()
            self.wt_init["components"]["monopile"]["internal_structure_2d_fem"]["outfitting_factor"] = float(
                wt_opt["monopile.outfitting_factor"]
            )
            for i in range(self.modeling_options["WISDEM"]["FixedBottomSE"]["n_layers_monopile"]):
                self.wt_init["components"]["monopile"]["internal_structure_2d_fem"]["layers"][i]["thickness"][
                    "grid"
                ] = wt_opt["monopile.s"].tolist()
                self.wt_init["components"]["monopile"]["internal_structure_2d_fem"]["layers"][i]["thickness"][
                    "values"
                ] = wt_opt["monopile.layer_thickness"][i, :].tolist()

        # Update jacket
        if self.modeling_options["flags"]["jacket"]:
            self.wt_init["components"]["jacket"]["r_head"] = float(wt_opt["jacket.r_head"][0])
            self.wt_init["components"]["jacket"]["r_foot"] = float(
                wt_opt["jacket.r_head"] * wt_opt["jacket.foot_head_ratio"]
            )
            self.wt_init["components"]["jacket"]["height"] = float(wt_opt["jacket.height"][0])
            self.wt_init["components"]["jacket"]["leg_diameter"] = float(wt_opt["jacket.leg_diameter"][0])
            self.wt_init["components"]["jacket"]["leg_thickness"] = float(wt_opt["jacket.leg_thickness"][0])
            self.wt_init["components"]["jacket"]["brace_diameters"] = wt_opt["jacket.brace_diameters"].tolist()
            self.wt_init["components"]["jacket"]["brace_thicknesses"] = wt_opt["jacket.brace_thicknesses"].tolist()
            self.wt_init["components"]["jacket"]["bay_spacing"] = wt_opt["jacket.bay_spacing"].tolist()

        # Update floating platform and mooring
        if self.modeling_options["flags"]["floating"]:
            yaml_out = self.wt_init["components"]["floating_platform"]
            n_joints = self.modeling_options["floating"]["joints"]["n_joints"]
            for i in range(n_joints):
                yaml_out["joints"][i]["location"] = wt_opt["floating.location"][i, :].tolist()

            n_members = self.modeling_options["floating"]["members"]["n_members"]
            for i in range(n_members):
                name_member = self.modeling_options["floating"]["members"]["name"][i]
                idx = self.modeling_options["floating"]["members"]["name2idx"][name_member]

                yaml_out["members"][i]["outer_shape"]["outer_diameter"]["grid"] = wt_opt[
                    f"floating.memgrp{idx}.s_in"
                ].tolist()
                yaml_out["members"][i]["outer_shape"]["outer_diameter"]["values"] = wt_opt[
                    f"floating.memgrp{idx}.outer_diameter_in"
                ].tolist()

                istruct = yaml_out["members"][i]["internal_structure"]

                n_layers = self.modeling_options["floating"]["members"]["n_layers"][i]
                for j in range(n_layers):
                    istruct["layers"][j]["thickness"]["grid"] = wt_opt[f"floating.memgrp{idx}.s_in"].tolist()
                    istruct["layers"][j]["thickness"]["values"] = wt_opt[f"floating.memgrp{idx}.layer_thickness_in"][
                        j, :
                    ].tolist()

                if "ring_stiffeners" in istruct:
                    istruct["ring_stiffeners"]["web_height"] = float(
                        wt_opt[f"floating.memgrp{idx}.ring_stiffener_web_height"]
                    )
                    istruct["ring_stiffeners"]["web_thickness"] = float(
                        wt_opt[f"floating.memgrp{idx}.ring_stiffener_web_thickness"]
                    )
                    istruct["ring_stiffeners"]["flange_thickness"] = float(
                        wt_opt[f"floating.memgrp{idx}.ring_stiffener_flange_thickness"]
                    )
                    istruct["ring_stiffeners"]["flange_width"] = float(
                        wt_opt[f"floating.memgrp{idx}.ring_stiffener_flange_width"]
                    )
                    istruct["ring_stiffeners"]["spacing"] = float(
                        wt_opt[f"floating.memgrp{idx}.ring_stiffener_spacing"]
                    )

                if "longitudinal_stiffeners" in istruct:
                    istruct["longitudinal_stiffeners"]["web_height"] = float(
                        wt_opt[f"floating.memgrp{idx}.axial_stiffener_web_height"]
                    )
                    istruct["longitudinal_stiffeners"]["web_thickness"] = float(
                        wt_opt[f"floating.memgrp{idx}.axial_stiffener_web_thickness"]
                    )
                    istruct["longitudinal_stiffeners"]["flange_thickness"] = float(
                        wt_opt[f"floating.memgrp{idx}.axial_stiffener_flange_thickness"]
                    )
                    istruct["longitudinal_stiffeners"]["flange_width"] = float(
                        wt_opt[f"floating.memgrp{idx}.axial_stiffener_flange_width"]
                    )
                    istruct["longitudinal_stiffeners"]["spacing"] = float(
                        wt_opt[f"floating.memgrp{idx}.axial_stiffener_spacing"]
                    )

                n_ballasts = self.modeling_options["floating"]["members"]["n_ballasts"][i]
                for j in range(n_ballasts):
                    if self.modeling_options["floating"]["members"]["ballast_flag_member_" + name_member][j] == False:
                        istruct["ballasts"][j]["volume"] = float(wt_opt[f"floating.memgrp{idx}.ballast_volume"][j])

                if self.modeling_options["floating"]["members"]["n_axial_joints"][i] > 0:
                    for j in range(self.modeling_options["floating"]["members"]["n_axial_joints"][i]):
                        yaml_out["members"][i]["axial_joints"][j]["grid"] = float(
                            wt_opt[f"floating.memgrp{idx}.grid_axial_joints"][j]
                        )

        if self.modeling_options["flags"]["mooring"]:
            n_lines = self.modeling_options["mooring"]["n_lines"]
            n_line_types = self.modeling_options["mooring"]["n_line_types"]
            line_names = [self.wt_init["components"]["mooring"]["line_types"][i]["name"] for i in range(n_line_types)]
            line_id = [self.wt_init["components"]["mooring"]["lines"][i]["line_type"] for i in range(n_lines)]

            for i in range(n_lines):
                self.wt_init["components"]["mooring"]["lines"][i]["unstretched_length"] = float(
                    wt_opt["mooring.unstretched_length"][i]
                )

            for jj, jname in enumerate(line_id):
                for ii, iname in enumerate(line_names):
                    if jname == iname:
                        self.wt_init["components"]["mooring"]["line_types"][ii]["diameter"] = float(
                            wt_opt["mooring.line_diameter"][jj]
                        )

        # Update rotor nacelle assembly
        if self.modeling_options["flags"]["RNA"]:
            self.wt_init["components"]["RNA"] = {}
            self.wt_init["components"]["RNA"]["elastic_properties_mb"] = {}
            self.wt_init["components"]["RNA"]["elastic_properties_mb"]["mass"] = float(wt_opt["drivese.rna_mass"])
            self.wt_init["components"]["RNA"]["elastic_properties_mb"]["inertia"] = wt_opt["drivese.rna_I_TT"].tolist()
            self.wt_init["components"]["RNA"]["elastic_properties_mb"]["center_mass"] = wt_opt[
                "drivese.rna_cm"
            ].tolist()

        # Update rotor diameter and hub height
        if self.modeling_options["flags"]["blade"]:
            self.wt_init["assembly"]["rotor_diameter"] = float(wt_opt["blade.high_level_blade_props.rotor_diameter"])
        self.wt_init["assembly"]["hub_height"] = float(wt_opt["high_level_tower_props.hub_height"])

        # Update controller
        if self.modeling_options["flags"]["control"]:
            self.wt_init["control"]["torque"]["tsr"] = float(wt_opt["control.rated_TSR"])

        # Write yamls with updated values
        sch.write_geometry_yaml(self.wt_init, fname_output)

    def write_options(self, fname_output):
        sch.write_modeling_yaml(self.modeling_options, fname_output)
        sch.write_analysis_yaml(self.analysis_options, fname_output)



import copy

import numpy as np
import pandas as pd
import openmdao.api as om
import wisdem.commonse.utilities as util

# -------------------------------------------------------------------------


class MainBearing(om.ExplicitComponent):
    """
    MainBearings class is used to represent the main bearing components of a wind turbine drivetrain.
    This is a simple, regression-based sizing tool for the main bearings.  The same function is called once
    for configurations with one main bearing or twice for configurations with two.  It handles Compact Aligning
    Roller Bearings (CARB), Cylindrical Roller Bearings (CRB), Spherical Roller Bearings (SRB), and
    Tapered Roller Bearings (TRB).

    Parameters
    ----------
    bearing_type : string
        bearing mass type
    D_bearing : float, [m]
        bearing diameter/facewidth
    D_shaft : float, [m]
        Diameter of LSS shaft at bearing location

    Returns
    -------
    mb_max_defl_ang : float, [rad]
        Maximum allowable deflection angle
    mb_mass : float, [kg]
        overall component mass
    mb_I : numpy array[3], [kg*m**2]
        moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass

    """

    def setup(self):
        self.add_discrete_input("bearing_type", "CARB")
        self.add_input("D_bearing", 0.0, units="m")
        self.add_input("D_shaft", 0.0, units="m")

        self.add_output("mb_max_defl_ang", 0.0, units="rad")
        self.add_output("mb_mass", 0.0, units="kg")
        self.add_output("mb_I", np.zeros(3), units="kg*m**2")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        if type(discrete_inputs["bearing_type"]) != type(""):
            raise ValueError("Bearing type input must be a string")
        btype = discrete_inputs["bearing_type"].upper()
        D_shaft = inputs["D_shaft"]

        # assume low load rating for bearing
        if btype == "CARB":  # p = Fr, so X=1, Y=0
            face_width = 0.2663 * D_shaft + 0.0435
            mass = 1561.4 * D_shaft ** 2.6007
            max_ang = np.deg2rad(0.5)

        elif btype == "CRB":
            face_width = 0.1136 * D_shaft
            mass = 304.19 * D_shaft ** 1.8885
            max_ang = np.deg2rad(4.0 / 60.0)

        elif btype == "SRB":
            face_width = 0.2762 * D_shaft
            mass = 876.7 * D_shaft ** 1.7195
            max_ang = 0.078

        # elif btype == 'RB':  # factors depend on ratio Fa/C0, C0 depends on bearing... TODO: add this functionality
        #    face_width = 0.0839
        #    mass = 229.47 * D_shaft**1.8036
        #    max_ang = 0.002

        # elif btype == 'TRB':
        #    face_width = 0.0740
        #    mass = 92.863 * D_shaft**.8399
        #    max_ang = np.deg2rad(3.0 / 60.0)

        elif btype == "TRB":
            face_width = 0.1499 * D_shaft
            mass = 543.01 * D_shaft ** 1.9043
            max_ang = np.deg2rad(3.0 / 60.0)

        else:
            raise ValueError("Bearing type must be CARB / CRB / SRB / TRB")

        # add housing weight, but pg 23 of report says factor is 2.92 whereas this is 2.963
        mass *= 1 + 80.0 / 27.0

        # Consider the bearings a torus for MoI (https://en.wikipedia.org/wiki/List_of_moments_of_inertia)
        D_bearing = inputs["D_bearing"] if inputs["D_bearing"] > 0.0 else face_width
        I0 = 0.25 * mass * (4 * (0.5 * D_shaft) ** 2 + 3 * (0.5 * D_bearing) ** 2)
        I1 = 0.125 * mass * (4 * (0.5 * D_shaft) ** 2 + 5 * (0.5 * D_bearing) ** 2)
        I = np.r_[I0, I1, I1]
        outputs["mb_mass"] = mass
        outputs["mb_I"] = I
        outputs["mb_max_defl_ang"] = max_ang


# -------------------------------------------------------------------


class Brake(om.ExplicitComponent):
    """
    The brake attaches to the high speed shaft for geared configurations or directly on the
    low speed shaft for direct drive configurations.  It is regression based, but also allows
    for a user override of the total mass value.

    Compute brake mass in the form of :math:`mass = k*torque`.
    Value of :math:`k` was updated in 2020 to be 0.00122.

    Parameters
    ----------
    rotor_diameter : float, [m]
        rotor diameter
    rated_torque : float, [N*m]
        rotor torque at rated power
    brake_mass_user : float, [kg]
        User override of brake mass
    D_shaft_end : float, [m]
        low speed shaft outer diameter
    s_rotor : float, [m]
        Generator rotor attachment to shaft s-coordinate
    s_gearbox : float, [m]
        Gearbox s-coordinate measured from bedplate
    rho : float, [kg/m**3]
        material density

    Returns
    -------
    brake_mass : float, [kg]
        overall component mass
    brake_cm : float, [m]
        center of mass of the component in [x,y,z] for an arbitrary coordinate system
    brake_I : numpy array[3], [kg*m**2]
        moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass
    """

    def initialize(self):
        self.options.declare("direct_drive", default=True)

    def setup(self):
        self.add_input("rotor_diameter", 0.0, units="m")
        self.add_input("rated_torque", 0.0, units="N*m")
        self.add_input("brake_mass_user", 0.0, units="kg")
        self.add_input("s_rotor", 0.0, units="m")
        self.add_input("s_gearbox", 0.0, units="m")

        self.add_output("brake_mass", 0.0, units="kg")
        self.add_output("brake_cm", 0.0, units="m")
        self.add_output("brake_I", np.zeros(3), units="kg*m**2")

    def compute(self, inputs, outputs):

        # Unpack inputs
        D_rotor = float(inputs["rotor_diameter"])
        Q_rotor = float(inputs["rated_torque"])
        m_brake = float(inputs["brake_mass_user"])
        s_rotor = float(inputs["s_rotor"])
        s_gearbox = float(inputs["s_gearbox"])

        # Regression based sizing derived by J.Keller under FOA 1981 support project
        if m_brake == 0.0:
            coeff = 0.00122
            m_brake = coeff * Q_rotor

        # Assume brake disc diameter and simple MoI
        D_disc = 0.01 * D_rotor
        Ib = np.zeros(3)
        Ib[0] = 0.5 * m_brake * (0.5 * D_disc) ** 2
        Ib[1:] = 0.5 * Ib[0]

        cm = s_rotor if self.options["direct_drive"] else 0.5 * (s_rotor + s_gearbox)

        outputs["brake_mass"] = m_brake
        outputs["brake_I"] = Ib
        outputs["brake_cm"] = cm


# ----------------------------------------------------------------------------------------------


class RPM_Input(om.ExplicitComponent):
    """
    Generates vector of possible rpm values from min to max.
    The max value is assumed to be the rated rpm value.

    Parameters
    ----------
    minimum_rpm : float, [rpm]
        Minimum shaft rotations-per-minute (rpm), usually set by controller
    rated_rpm : float, [rpm]
        Rated shaft rotations-per-minute (rpm)
    gear_ratio : float
        overall gearbox ratio

    Returns
    -------
    shaft_rpm : float, [rpm]
        Vector of possible rpm values
    """

    def initialize(self):
        self.options.declare("n_pc", default=20)

    def setup(self):
        n_pc = self.options["n_pc"]
        self.add_input("minimum_rpm", val=1.0, units="rpm")
        self.add_input("rated_rpm", val=0.0, units="rpm")
        self.add_input("gear_ratio", val=1.0)

        self.add_output("lss_rpm", val=np.zeros(n_pc), units="rpm")
        self.add_output("hss_rpm", val=np.zeros(n_pc), units="rpm")

    def compute(self, inputs, outputs):
        min_rpm = np.maximum(0.1, float(inputs["minimum_rpm"]))
        max_rpm = float(inputs["rated_rpm"])
        ratio = float(inputs["gear_ratio"])
        rpm_full = np.linspace(min_rpm, max_rpm, self.options["n_pc"])
        outputs["lss_rpm"] = rpm_full
        outputs["hss_rpm"] = ratio * rpm_full


# ----------------------------------------------------------------------------------------------


class GeneratorSimple(om.ExplicitComponent):
    """
    The Generator class is used to represent the generator of a wind turbine drivetrain
    using simple scaling laws.  For a more detailed electromagnetic and structural design,
    please see the other generator components.

    Parameters
    ----------
    rotor_diameter : float, [m]
        rotor diameter
    machine_rating : float, [kW]
        machine rating of generator
    L_generator : float, [m]
        Generator stack width
    rated_torque : float, [N*m]
        rotor torque at rated power
    shaft_rpm : numpy array[n_pc], [rpm]
        Input shaft rotations-per-minute (rpm)
    generator_mass_user : float, [kg]
        User input override of generator mass
    generator_radius_user : float, [m]
        User input override of generator radius

    Returns
    -------
    R_generator : float, [m]
        center of mass of the component in [x,y,z] for an arbitrary coordinate system
    generator_mass : float, [kg]
        overall component mass
    generator_I : numpy array[3], [kg*m**2]
        moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass
    generator_efficiency : numpy array[n_pc]
        Generator efficiency at various rpm values
    lss_rpm : numpy array[n_pc], [rpm]
        Low speed shaft rpm values at which the generator efficiency is given
    """

    def initialize(self):
        self.options.declare("direct_drive", default=True)
        self.options.declare("n_pc", default=20)

    def setup(self):
        n_pc = self.options["n_pc"]

        # variables
        self.add_input("rotor_diameter", val=0.0, units="m")
        self.add_input("machine_rating", val=0.0, units="kW")
        self.add_input("rated_torque", 0.0, units="N*m")
        self.add_input("lss_rpm", np.zeros(n_pc), units="rpm")
        self.add_input("L_generator", 0.0, units="m")
        self.add_input("generator_mass_user", 0.0, units="kg")
        self.add_input("generator_radius_user", 0.0, units="m")
        self.add_input("generator_efficiency_user", val=np.zeros((n_pc, 2)))

        self.add_output("R_generator", val=0.0, units="m")
        self.add_output("generator_mass", val=0.0, units="kg")
        self.add_output("generator_rotor_mass", val=0.0, units="kg")
        self.add_output("generator_stator_mass", val=0.0, units="kg")
        self.add_output("generator_rotor_I", val=np.zeros(3), units="kg*m**2")
        self.add_output("generator_stator_I", val=np.zeros(3), units="kg*m**2")
        self.add_output("generator_I", val=np.zeros(3), units="kg*m**2")
        self.add_output("generator_efficiency", val=np.zeros(n_pc))

    def compute(self, inputs, outputs):

        # Unpack inputs
        rating = float(inputs["machine_rating"])
        D_rotor = float(inputs["rotor_diameter"])
        Q_rotor = float(inputs["rated_torque"])
        R_generator = float(inputs["generator_radius_user"])
        mass = float(inputs["generator_mass_user"])
        eff_user = inputs["generator_efficiency_user"]
        length = float(inputs["L_generator"])

        if mass == 0.0:
            if self.options["direct_drive"]:
                massCoeff = 1e-3 * 37.68
                mass = massCoeff * Q_rotor
            else:
                massCoeff = np.mean([6.4737, 10.51, 5.34])
                massExp = 0.9223
                mass = massCoeff * rating ** massExp
        outputs["generator_mass"] = mass
        outputs["generator_rotor_mass"] = outputs["generator_stator_mass"] = 0.5 * mass

        # calculate mass properties
        if R_generator == 0.0:
            R_generator = 0.5 * 0.015 * D_rotor
        outputs["R_generator"] = R_generator

        I = np.zeros(3)
        I[0] = 0.5 * R_generator ** 2
        I[1:] = (1.0 / 12.0) * (3 * R_generator ** 2 + length ** 2)
        outputs["generator_I"] = mass * I
        outputs["generator_rotor_I"] = outputs["generator_stator_I"] = 0.5 * mass * I

        # Efficiency performance- borrowed and adapted from servose
        # Note: Have to use lss_rpm no matter what here because servose interpolation based on lss shaft rpm
        rpm_full = inputs["lss_rpm"]
        if np.any(eff_user):
            eff = np.interp(rpm_full, eff_user[:, 0], eff_user[:, 1])

        else:
            if self.options["direct_drive"]:
                constant = 0.01007
                linear = 0.02000
                quadratic = 0.06899
            else:
                constant = 0.01289
                linear = 0.08510
                quadratic = 0.0

            # Normalize by rated
            ratio = rpm_full / rpm_full[-1]
            eff = 1.0 - (constant / ratio + linear + quadratic * ratio)

        eff = np.maximum(1e-3, eff)
        outputs["generator_efficiency"] = eff


# -------------------------------------------------------------------------------


class Electronics(om.ExplicitComponent):
    """
    Estimate mass of electronics based on rating, rotor diameter, and tower top diameter.
    Empirical only, no load analysis.

    Parameters
    ----------
    machine_rating : float, [kW]
        machine rating of the turbine
    rotor_diameter : float, [m]
        rotor diameter of turbine
    D_top : float, [m]
        Tower top outer diameter
    converter_mass_user : float, [kg]
        Override regular regression-based calculation of converter mass with this value
    transformer_mass_user : float, [kg]
        Override regular regression-based calculation of transformer mass with this value

    Returns
    -------
    converter_mass : float, [kg]
        overall component mass
    converter_cm : numpy array[3], [m]
        center of mass of the component in [x,y,z] for an arbitrary coordinate system
    converter_I : numpy array[3], [kg*m**2]
        moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass
    transformer_mass : float, [kg]
        overall component mass
    transformer_cm : numpy array[3], [m]
        center of mass of the component in [x,y,z] for an arbitrary coordinate system
    transformer_I : numpy array[3], [kg*m**2]
        moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass

    """

    def setup(self):
        self.add_input("machine_rating", 0.0, units="kW")
        self.add_input("rotor_diameter", 0.0, units="m")
        self.add_input("D_top", 0.0, units="m")
        self.add_input("converter_mass_user", 0.0, units="kg")
        self.add_input("transformer_mass_user", 0.0, units="kg")

        self.add_output("converter_mass", 0.0, units="kg")
        self.add_output("converter_cm", np.zeros(3), units="m")
        self.add_output("converter_I", np.zeros(3), units="kg*m**2")
        self.add_output("transformer_mass", 0.0, units="kg")
        self.add_output("transformer_cm", np.zeros(3), units="m")
        self.add_output("transformer_I", np.zeros(3), units="kg*m**2")

    def compute(self, inputs, outputs):

        # Unpack inputs
        rating = float(inputs["machine_rating"])
        D_rotor = float(inputs["rotor_diameter"])
        D_top = float(inputs["D_top"])
        m_conv_usr = float(inputs["converter_mass_user"])
        m_trans_usr = float(inputs["transformer_mass_user"])

        # Correlation based trends, assume box
        m_converter = (
            m_conv_usr if m_conv_usr > 0.0 else 0.77875 * rating + 302.6
        )  # coeffs=1e-3*np.mean([740., 817.5]), np.mean([101.37, 503.83])
        m_transformer = m_trans_usr if m_trans_usr > 0.0 else 1.915 * rating + 1910.0

        # CM location, just assume off to the side of the bedplate
        cm = np.zeros(3)
        sides = 0.015 * D_rotor
        cm[1] = 0.5 * D_top + 0.5 * sides
        cm[2] = 0.5 * sides

        # Outputs
        outputs["converter_mass"] = m_converter
        outputs["converter_cm"] = cm
        outputs["converter_I"] = (1.0 / 6.0) * m_converter * sides ** 2 * np.ones(3)

        outputs["transformer_mass"] = m_transformer
        outputs["transformer_cm"] = cm
        outputs["transformer_cm"][1] *= -1.0
        outputs["transformer_I"] = (1.0 / 6.0) * m_transformer * sides ** 2 * np.ones(3)


# ---------------------------------------------------------------------------------------------------------------


class YawSystem(om.ExplicitComponent):
    """
    Estimate mass of yaw system based on rotor diameter and tower top diameter.
    Empirical only, no load analysis.

    Parameters
    ----------
    rotor_diameter : float, [m]
        rotor diameter
    D_top : float, [m]
        Tower top outer diameter
    rho : float, [kg/m**3]
        material density

    Returns
    -------
    yaw_mass : float, [kg]
        overall component mass
    yaw_cm : numpy array[3], [m]
        center of mass of the component in [x,y,z] for an arbitrary coordinate system
    yaw_I : numpy array[3], [kg*m**2]
        moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass

    """

    def setup(self):
        # variables
        self.add_input("rotor_diameter", 0.0, units="m")
        self.add_input("D_top", 0.0, units="m")
        self.add_input("rho", 0.0, units="kg/m**3")

        self.add_output("yaw_mass", 0.0, units="kg")
        self.add_output("yaw_cm", np.zeros(3), units="m")
        self.add_output("yaw_I", np.zeros(3), units="kg*m**2")

    def compute(self, inputs, outputs):
        # Unpack inputs
        D_rotor = float(inputs["rotor_diameter"])
        D_top = float(inputs["D_top"])
        rho = float(inputs["rho"])

        # Estimate the number of yaw motors (borrowed from old DriveSE utilities)
        n_motors = 2 * np.ceil(D_rotor / 30.0) - 2

        # Assume same yaw motors as Vestas V80 for now: Bonfiglioli 709T2M
        m_motor = 190.0

        # Assume friction plate surface width is 1/10 the diameter and thickness scales with rotor diameter
        m_frictionPlate = rho * np.pi * D_top * (0.1 * D_top) * (1e-3 * D_rotor)

        # Total mass estimate
        outputs["yaw_mass"] = m_frictionPlate + n_motors * m_motor

        # Assume cm is at tower top (cm=0,0,0) and mass is non-rotating (I=0,..), so leave at default value of 0s
        outputs["yaw_cm"] = np.zeros(3)
        outputs["yaw_I"] = np.zeros(3)


# ---------------------------------------------------------------------------------------------------------------


class MiscNacelleComponents(om.ExplicitComponent):
    """
    Estimate mass properties of miscellaneous other ancillary components in the nacelle.

    Parameters
    ----------
    upwind : boolean
        Flag whether the design is upwind or downwind
    machine_rating : float, [kW]
        machine rating of the turbine
    hvac_mass_coeff : float, [kg/kW]
        Regression-based scaling coefficient on machine rating to get HVAC system mass
    H_bedplate : float, [m]
        height of bedplate
    D_top : float, [m]
        Tower top outer diameter
    bedplate_mass : float, [kg]
        Bedplate mass
    bedplate_I : numpy array[6], [kg*m**2]
        Bedplate mass moment of inertia about base
    R_generator : float, [m]
        Generatour outer diameter
    overhang : float, [m]
        Overhang of rotor from tower along x-axis in yaw-aligned c.s.
    cm_generator : float, [m]
        Generator center of mass s-coordinate
    rho_fiberglass : float, [kg/m**3]
        material density of fiberglass

    Returns
    -------
    hvac_mass : float, [kg]
        component mass
    hvac_cm : float, [m]
        component center of mass
    hvac_I : numpy array[3], [m]
        component mass moments of inertia
    platform_mass : float, [kg]
        component mass
    platform_cm : numpy array[3], [m]
        component center of mass
    platform_I : numpy array[3], [m]
        component mass moments of inertia
    cover_mass : float, [kg]
        component mass
    cover_cm : numpy array[3], [m]
        component center of mass
    cover_I : numpy array[3], [m]
        component mass moments of inertia

    """

    def initialize(self):
        self.options.declare("direct_drive", default=True)

    def setup(self):
        self.add_discrete_input("upwind", True)
        self.add_input("machine_rating", 0.0, units="kW")
        self.add_input("hvac_mass_coeff", 0.025, units="kg/kW/m")
        self.add_input("H_bedplate", 0.0, units="m")
        self.add_input("D_top", 0.0, units="m")
        self.add_input("L_bedplate", 0.0, units="m")
        self.add_input("R_generator", 0.0, units="m")
        self.add_input("overhang", 0.0, units="m")
        self.add_input("generator_cm", 0.0, units="m")
        self.add_input("rho_fiberglass", 0.0, units="kg/m**3")
        self.add_input("rho_castiron", 0.0, units="kg/m**3")

        self.add_output("hvac_mass", 0.0, units="kg")
        self.add_output("hvac_cm", 0.0, units="m")
        self.add_output("hvac_I", np.zeros(3), units="m")
        self.add_output("platform_mass", 0.0, units="kg")
        self.add_output("platform_cm", np.zeros(3), units="m")
        self.add_output("platform_I", np.zeros(3), units="m")
        self.add_output("cover_mass", 0.0, units="kg")
        self.add_output("cover_cm", np.zeros(3), units="m")
        self.add_output("cover_I", np.zeros(3), units="m")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        # Unpack inputs
        direct = self.options["direct_drive"]
        upwind = discrete_inputs["upwind"]
        rating = float(inputs["machine_rating"])
        coeff = float(inputs["hvac_mass_coeff"])
        H_bedplate = float(inputs["H_bedplate"])
        D_top = float(inputs["D_top"])
        L_bedplate = float(inputs["L_bedplate"])
        R_generator = float(inputs["R_generator"])
        overhang = float(inputs["overhang"])
        s_generator = float(inputs["generator_cm"])
        rho_fiberglass = float(inputs["rho_fiberglass"])
        rho_castiron = float(inputs["rho_castiron"])

        # For the nacelle cover, imagine a box from the bedplate to the hub in length and around the generator in width, height, with 10% margin in each dim
        L_cover = 1.1 * L_bedplate if direct else 1.1 * (overhang + D_top)
        W_cover = 1.1 * 2 * R_generator
        H_cover = 1.1 * (R_generator + np.maximum(R_generator, H_bedplate))
        A_cover = 2 * (L_cover * W_cover + L_cover * H_cover + H_cover * W_cover)
        t_cover = 0.02
        m_cover = A_cover * t_cover * rho_fiberglass
        cm_cover = np.array([0.5 * L_cover - 0.5 * L_bedplate, 0.0, 0.5 * H_cover])
        I_cover = (
            m_cover
            * np.array(
                [
                    H_cover ** 2 + W_cover ** 2 - (H_cover - t_cover) ** 2 - (W_cover - t_cover) ** 2,
                    H_cover ** 2 + L_cover ** 2 - (H_cover - t_cover) ** 2 - (L_cover - t_cover) ** 2,
                    W_cover ** 2 + L_cover ** 2 - (W_cover - t_cover) ** 2 - (L_cover - t_cover) ** 2,
                ]
            )
            / 12.0
        )
        if upwind:
            cm_cover[0] *= -1.0
        outputs["cover_mass"] = m_cover
        outputs["cover_cm"] = cm_cover
        outputs["cover_I"] = I_cover

        # Regression based estimate on HVAC mass
        m_hvac = coeff * rating * 2 * np.pi * (0.75 * R_generator)
        cm_hvac = s_generator
        I_hvac = m_hvac * (0.75 * R_generator) ** 2
        outputs["hvac_mass"] = m_hvac
        outputs["hvac_cm"] = cm_hvac
        outputs["hvac_I"] = I_hvac * np.array([1.0, 0.5, 0.5])

        # Platforms as a fraction of bedplate mass and bundling it to call it 'platforms'
        L_platform = 2 * D_top if direct else L_cover
        W_platform = 2 * D_top if direct else W_cover
        t_platform = 0.04
        m_platform = L_platform * W_platform * t_platform * rho_castiron
        I_platform = (
            m_platform
            * np.array(
                [
                    t_platform ** 2 + W_platform ** 2,
                    t_platform ** 2 + L_platform ** 2,
                    W_platform ** 2 + L_platform ** 2,
                ]
            )
            / 12.0
        )
        outputs["platform_mass"] = m_platform
        outputs["platform_cm"] = np.zeros(3)
        outputs["platform_I"] = I_platform


# --------------------------------------------
class NacelleSystemAdder(om.ExplicitComponent):  # added to drive to include electronics
    """
    The Nacelle class is used to represent the overall nacelle of a wind turbine.

    Parameters
    ----------
    upwind : boolean
        Flag whether the design is upwind or downwind
    uptower : boolean
        Power electronics are placed in the nacelle at the tower top
    tilt : float, [deg]
        Shaft tilt
    mb1_mass : float, [kg]
        component mass
    mb1_cm : float, [m]
        component CM
    mb1_I : numpy array[3], [kg*m**2]
        component I
    mb2_mass : float, [kg]
        component mass
    mb2_cm : float, [m]
        component CM
    mb2_I : numpy array[3], [kg*m**2]
        component I
    gearbox_mass : float, [kg]
        component mass
    gearbox_cm : numpy array[3], [m]
        component CM
    gearbox_I : numpy array[3], [kg*m**2]
        component I
    hss_mass : float, [kg]
        component mass
    hss_cm : float, [m]
        component CM
    hss_I : numpy array[3], [kg*m**2]
        component I
    brake_mass : float, [kg]
        component mass
    brake_cm : float, [m]
        component CM
    brake_I : numpy array[3], [kg*m**2]
        component I
    generator_mass : float, [kg]
        component mass
    generator_cm : float, [m]
        component CM
    generator_I : numpy array[3], [kg*m**2]
        component I
    nose_mass : float, [kg]
        Nose mass
    nose_cm : float, [m]
        Nose center of mass along nose axis from bedplate
    nose_I : numpy array[3], [kg*m**2]
        Nose moment of inertia around cm in axial (hub-aligned) c.s.
    lss_mass : float, [kg]
        LSS mass
    lss_cm : float, [m]
        LSS center of mass along shaft axis from bedplate
    lss_I : numpy array[3], [kg*m**2]
        LSS moment of inertia around cm in axial (hub-aligned) c.s.
    converter_mass : float, [kg]
        overall component mass
    converter_cm : numpy array[3], [m]
        center of mass of the component in [x,y,z] for an arbitrary coordinate system
    converter_I : numpy array[3], [kg*m**2]
        moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass
    transformer_mass : float, [kg]
        overall component mass
    transformer_cm : numpy array[3], [m]
        center of mass of the component in [x,y,z] for an arbitrary coordinate system
    transformer_I : numpy array[3], [kg*m**2]
        moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass
    yaw_mass : float, [kg]
        overall component mass
    yaw_cm : numpy array[3], [m]
        center of mass of the component in [x,y,z] for an arbitrary coordinate system
    yaw_I : numpy array[3], [kg*m**2]
        moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass
    bedplate_mass : float, [kg]
        component mass
    bedplate_cm : numpy array[3], [m]
        component CM
    bedplate_I : numpy array[6], [kg*m**2]
        component I
    hvac_mass : float, [kg]
        component mass
    hvac_cm : float, [m]
        component center of mass
    hvac_I : numpy array[3], [m]
        component mass moments of inertia
    platform_mass : float, [kg]
        component mass
    platform_cm : numpy array[3], [m]
        component center of mass
    platform_I : numpy array[3], [m]
        component mass moments of inertia
    cover_mass : float, [kg]
        component mass
    cover_cm : numpy array[3], [m]
        component center of mass
    cover_I : numpy array[3], [m]
        component mass moments of inertia

    Returns
    -------
    shaft_start : numpy array[3], [m]
        coordinate of start of shaft relative to tower top
    other_mass : float, [kg]
        mass of high speed shaft, hvac, main frame, yaw, cover, and electronics
    mean_bearing_mass : float, [kg]
        average mass of all bearings (currently 2) for summary mass and cost calculations
    total_bedplate_mass : float, [kg]
        bedplate and nose mass for summary mass and cost calculations
    above_yaw_mass : float, [kg]
       overall nacelle mass excluding yaw
    nacelle_mass : float, [kg]
        overall nacelle mass including yaw
    nacelle_cm : numpy array[3], [m]
        coordinates of the center of mass of the nacelle (including yaw) in tower top coordinate system [x,y,z]
    above_yaw_cm : numpy array[3], [m]
        coordinates of the center of mass of the nacelle (excluding yaw) in tower top coordinate system [x,y,z]
    nacelle_I : numpy array[6], [kg*m**2]
        moments of inertia for the nacelle (including yaw) [Ixx, Iyy, Izz, Ixy, Ixz, Iyz] around its center of mass
    nacelle_I_TT : numpy array[6], [kg*m**2]
        moments of inertia for the nacelle (including yaw) [Ixx, Iyy, Izz, Ixy, Ixz, Iyz] around the tower top
    above_yaw_I : numpy array[6], [kg*m**2]
        moments of inertia for the nacelle (excluding yaw) [Ixx, Iyy, Izz, Ixy, Ixz, Iyz] around its center of mass
    above_yaw_I_TT : numpy array[6], [kg*m**2]
        moments of inertia for the nacelle (excluding yaw) [Ixx, Iyy, Izz, Ixy, Ixz, Iyz] around the tower top
    """

    def initialize(self):
        self.options.declare("direct_drive", default=True)

    def setup(self):
        self.add_discrete_input("upwind", True)
        self.add_discrete_input("uptower", True)
        self.add_input("tilt", 0.0, units="deg")
        self.add_input("mb1_mass", 0.0, units="kg")
        self.add_input("mb1_cm", 0.0, units="m")
        self.add_input("mb1_I", np.zeros(3), units="kg*m**2")
        self.add_input("mb2_mass", 0.0, units="kg")
        self.add_input("mb2_cm", 0.0, units="m")
        self.add_input("mb2_I", np.zeros(3), units="kg*m**2")
        self.add_input("gearbox_mass", 0.0, units="kg")
        self.add_input("gearbox_cm", 0.0, units="m")
        self.add_input("gearbox_I", np.zeros(3), units="kg*m**2")
        self.add_input("hss_mass", 0.0, units="kg")
        self.add_input("hss_cm", 0.0, units="m")
        self.add_input("hss_I", np.zeros(3), units="kg*m**2")
        self.add_input("brake_mass", 0.0, units="kg")
        self.add_input("brake_cm", 0.0, units="m")
        self.add_input("brake_I", np.zeros(3), units="kg*m**2")
        self.add_input("generator_mass", 0.0, units="kg")
        self.add_input("generator_rotor_mass", val=0.0, units="kg")
        self.add_input("generator_stator_mass", val=0.0, units="kg")
        self.add_input("generator_cm", 0.0, units="m")
        self.add_input("generator_I", np.zeros(3), units="kg*m**2")
        self.add_input("generator_rotor_I", val=np.zeros(3), units="kg*m**2")
        self.add_input("generator_stator_I", val=np.zeros(3), units="kg*m**2")
        self.add_input("nose_mass", val=0.0, units="kg")
        self.add_input("nose_cm", val=0.0, units="m")
        self.add_input("nose_I", val=np.zeros(3), units="kg*m**2")
        self.add_input("lss_mass", val=0.0, units="kg")
        self.add_input("lss_cm", val=0.0, units="m")
        self.add_input("lss_I", val=np.zeros(3), units="kg*m**2")
        self.add_input("converter_mass", 0.0, units="kg")
        self.add_input("converter_cm", np.zeros(3), units="m")
        self.add_input("converter_I", np.zeros(3), units="kg*m**2")
        self.add_input("transformer_mass", 0.0, units="kg")
        self.add_input("transformer_cm", np.zeros(3), units="m")
        self.add_input("transformer_I", np.zeros(3), units="kg*m**2")
        self.add_input("yaw_mass", 0.0, units="kg")
        self.add_input("yaw_cm", np.zeros(3), units="m")
        self.add_input("yaw_I", np.zeros(3), units="kg*m**2")
        self.add_input("bedplate_mass", 0.0, units="kg")
        self.add_input("bedplate_cm", np.zeros(3), units="m")
        self.add_input("bedplate_I", np.zeros(6), units="kg*m**2")
        self.add_input("hvac_mass", 0.0, units="kg")
        self.add_input("hvac_cm", 0.0, units="m")
        self.add_input("hvac_I", np.zeros(3), units="m")
        self.add_input("platform_mass", 0.0, units="kg")
        self.add_input("platform_cm", np.zeros(3), units="m")
        self.add_input("platform_I", np.zeros(3), units="m")
        self.add_input("cover_mass", 0.0, units="kg")
        self.add_input("cover_cm", np.zeros(3), units="m")
        self.add_input("cover_I", np.zeros(3), units="m")
        self.add_input("x_bedplate", val=np.zeros(12), units="m")
        self.add_input("constr_height", 0.0, units="m")

        self.add_output("shaft_start", np.zeros(3), units="m")
        self.add_output("other_mass", 0.0, units="kg")
        self.add_output("mean_bearing_mass", 0.0, units="kg")
        self.add_output("total_bedplate_mass", 0.0, units="kg")
        self.add_output("nacelle_mass", 0.0, units="kg")
        self.add_output("above_yaw_mass", 0.0, units="kg")
        self.add_output("nacelle_cm", np.zeros(3), units="m")
        self.add_output("above_yaw_cm", np.zeros(3), units="m")
        self.add_output("nacelle_I", np.zeros(6), units="kg*m**2")
        self.add_output("nacelle_I_TT", np.zeros(6), units="kg*m**2")
        self.add_output("above_yaw_I", np.zeros(6), units="kg*m**2")
        self.add_output("above_yaw_I_TT", np.zeros(6), units="kg*m**2")

        self._mass_table = None

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        Cup = -1.0 if discrete_inputs["upwind"] else 1.0
        tilt = float(np.deg2rad(inputs["tilt"]))

        components = [
            "mb1",
            "mb2",
            "lss",
            "hss",
            "brake",
            "gearbox",
            "generator_rotor",
            "generator_stator",
            "generator",
            "hvac",
            "nose",
            "bedplate",
            "platform",
            "cover",
        ]
        if discrete_inputs["uptower"]:
            components.extend(["transformer", "converter"])

        # Mass and CofM summaries first because will need them for I later
        m_nac = 0.0
        cm_nac = np.zeros(3)
        shaft0 = np.zeros(3)
        shaft0[-1] += inputs["constr_height"]
        if self.options["direct_drive"]:
            shaft0[0] += inputs["x_bedplate"][-1]
        outputs["shaft_start"] = shaft0

        for k in components:
            if k in ["generator_rotor", "generator_stator"]:
                continue
            m_i = inputs[k + "_mass"]
            cm_i = inputs[k + "_cm"]

            # If cm is (x,y,z) then it is already in tower-top c.s.  If it is a scalar, it is in distance from tower and we have to convert
            if len(cm_i) == 1:
                cm_i = shaft0 + cm_i * np.array([Cup * np.cos(tilt), 0.0, np.sin(tilt)])

            m_nac += m_i
            cm_nac += m_i * cm_i

        # Complete CofM calculation
        cm_nac /= m_nac

        # Now find total I about nacelle CofM
        I_nac = np.zeros(6)
        m_list = np.zeros((len(components) + 3,))
        cm_list = np.zeros((len(components) + 3, 3))
        I_cm_list = np.zeros((len(components) + 3, 6))
        I_TT_list = np.zeros((len(components) + 3, 6))
        for ic, c in enumerate(components):
            m_i = inputs[c + "_mass"]
            cm_i = inputs["generator_cm"] if c.find("generator") >= 0 else inputs[c + "_cm"]
            I_i = inputs[c + "_I"]

            # Rotate MofI if in hub c.s.
            if len(cm_i) == 1:
                cm_i = shaft0 + cm_i * np.array([Cup * np.cos(tilt), 0.0, np.sin(tilt)])
                I_i = util.rotateI(I_i, -Cup * tilt, axis="y")
            else:
                I_i = np.r_[I_i, np.zeros(3)]

            r = cm_i - cm_nac
            if not c in ["generator_rotor", "generator_stator"]:
                I_add = util.assembleI(I_i) + m_i * (np.dot(r, r) * np.eye(3) - np.outer(r, r))
                I_add = util.unassembleI(I_add)
                I_nac += I_add

            # Record mass, cm, and I for output table
            m_list[ic] = m_i
            cm_list[ic, :] = cm_i
            I_TT_list[ic, :] = util.unassembleI(
                util.assembleI(I_i) + m_i * (np.dot(cm_i, cm_i) * np.eye(3) - np.outer(cm_i, cm_i))
            )
            I_i = inputs[c + "_I"]
            I_cm_list[ic, :] = I_i if I_i.size == 6 else np.r_[I_i, np.zeros(3)]

        outputs["above_yaw_mass"] = copy.copy(m_nac)
        outputs["above_yaw_cm"] = R = cm_nac.copy()
        outputs["above_yaw_I"] = I_nac.copy()
        parallel_axis = m_nac * (np.dot(R, R) * np.eye(3) - np.outer(R, R))
        outputs["above_yaw_I_TT"] = util.unassembleI(util.assembleI(I_nac) + parallel_axis)

        m_nac += inputs["yaw_mass"]
        cm_nac = (outputs["above_yaw_mass"] * outputs["above_yaw_cm"] + inputs["yaw_cm"] * inputs["yaw_mass"]) / m_nac
        r = inputs["yaw_cm"] - cm_nac
        I_add = util.assembleI(np.r_[inputs["yaw_I"], np.zeros(3)]) + inputs["yaw_mass"] * (
            np.dot(r, r) * np.eye(3) - np.outer(r, r)
        )
        I_add = util.unassembleI(I_add)
        I_nac += I_add

        outputs["nacelle_mass"] = m_nac
        outputs["nacelle_cm"] = cm_nac
        outputs["nacelle_I"] = I_nac

        # Find nacelle MoI about tower top
        R = cm_nac
        parallel_axis = m_nac * (np.dot(R, R) * np.eye(3) - np.outer(R, R))
        outputs["nacelle_I_TT"] = util.unassembleI(util.assembleI(I_nac) + parallel_axis)

        # Store other misc outputs
        outputs["other_mass"] = (
            inputs["hvac_mass"]
            + inputs["platform_mass"]
            + inputs["cover_mass"]
            + inputs["yaw_mass"]
            + inputs["converter_mass"]
            + inputs["transformer_mass"]
        )
        outputs["mean_bearing_mass"] = 0.5 * (inputs["mb1_mass"] + inputs["mb2_mass"])
        outputs["total_bedplate_mass"] = inputs["nose_mass"] + inputs["bedplate_mass"]

        # Wrap up nacelle mass table
        components.append("Above_yaw")
        m_list[-3] = outputs["above_yaw_mass"]
        cm_list[-3, :] = outputs["above_yaw_cm"]
        I_cm_list[-3, :] = outputs["above_yaw_I"]
        I_TT_list[-3, :] = outputs["above_yaw_I_TT"]
        components.append("yaw")
        m_list[-2] = inputs["yaw_mass"]
        cm_list[-2, :] = inputs["yaw_cm"]
        I_cm_list[-2, :] = I_TT_list[-2, :] = np.r_[inputs["yaw_I"], np.zeros(3)]
        components.append("nacelle")
        m_list[-1] = m_nac
        cm_list[-1, :] = cm_nac
        I_cm_list[-1, :] = I_nac
        I_TT_list[-1, :] = outputs["nacelle_I_TT"]
        self._mass_table = pd.DataFrame()
        self._mass_table["Component"] = components
        self._mass_table["Mass"] = m_list
        self._mass_table["CoM_TT_x"] = cm_list[:, 0]
        self._mass_table["CoM_TT_y"] = cm_list[:, 1]
        self._mass_table["CoM_TT_z"] = cm_list[:, 2]
        self._mass_table["MoI_CoM_xx"] = I_cm_list[:, 0]
        self._mass_table["MoI_CoM_yy"] = I_cm_list[:, 1]
        self._mass_table["MoI_CoM_zz"] = I_cm_list[:, 2]
        self._mass_table["MoI_CoM_xy"] = I_cm_list[:, 3]
        self._mass_table["MoI_CoM_xz"] = I_cm_list[:, 4]
        self._mass_table["MoI_CoM_yz"] = I_cm_list[:, 5]
        self._mass_table["MoI_TT_xx"] = I_TT_list[:, 0]
        self._mass_table["MoI_TT_yy"] = I_TT_list[:, 1]
        self._mass_table["MoI_TT_zz"] = I_TT_list[:, 2]
        self._mass_table["MoI_TT_xy"] = I_TT_list[:, 3]
        self._mass_table["MoI_TT_xz"] = I_TT_list[:, 4]
        self._mass_table["MoI_TT_yz"] = I_TT_list[:, 5]
        self._mass_table.set_index("Component", inplace=True)


# --------------------------------------------


class RNA_Adder(om.ExplicitComponent):
    """
    Compute mass and moments of inertia for RNA system.

    Parameters
    ----------
    upwind : boolean
        Flag whether the design is upwind or downwind
    tilt : float, [deg]
        Shaft tilt
    L_drive : float, [m]
        Length of drivetrain from bedplate to hub flang
    shaft_start : numpy array[3], [m]
        coordinate of start of shaft relative to tower top
    blades_mass : float, [kg]
        Mass of all bladea
    hub_system_mass : float, [kg]
        Mass of hub system (hub + spinner + pitch)
    nacelle_mass : float, [kg]
        Mass of nacelle system
    hub_system_cm : float, [m]
        Hub center of mass from hub flange in hub c.s.
    nacelle_cm : numpy array[3], [m]
        Nacelle center of mass relative to tower top in yaw-aligned c.s.
    blades_I : numpy array[6], [kg*m**2]
        Mass moments of inertia of all blades about hub center
    hub_system_I : numpy array[6], [kg*m**2]
        Mass moments of inertia of hub system about its CofM
    nacelle_I_TT : numpy array[6], [kg*m**2]
        Mass moments of inertia of nacelle about the tower top

    Returns
    -------
    rotor_mass : float, [kg]
        Mass of blades and hub system
    rna_mass : float, [kg]
        Total RNA mass
    rna_cm : numpy array[3], [m]
        RNA center of mass relative to tower top in yaw-aligned c.s.
    rna_I_TT : numpy array[6], [kg*m**2]
        Mass moments of inertia of RNA about tower top in yaw-aligned coordinate system
    """

    def setup(self):

        self.add_discrete_input("upwind", True)
        self.add_input("tilt", 0.0, units="deg")
        self.add_input("L_drive", 0.0, units="m")
        self.add_input("shaft_start", np.zeros(3), units="m")
        self.add_input("blades_mass", 0.0, units="kg")
        self.add_input("hub_system_mass", 0.0, units="kg")
        self.add_input("nacelle_mass", 0.0, units="kg")
        self.add_input("hub_system_cm", 0.0, units="m")
        self.add_input("nacelle_cm", np.zeros(3), units="m")
        self.add_input("blades_I", np.zeros(6), units="kg*m**2")
        self.add_input("hub_system_I", np.zeros(6), units="kg*m**2")
        self.add_input("nacelle_I_TT", np.zeros(6), units="kg*m**2")

        self.add_output("rotor_mass", 0.0, units="kg")
        self.add_output("rna_mass", 0.0, units="kg")
        self.add_output("rna_cm", np.zeros(3), units="m")
        self.add_output("rna_I_TT", np.zeros(6), units="kg*m**2")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        Cup = -1.0 if discrete_inputs["upwind"] else 1.0
        tilt = float(np.deg2rad(inputs["tilt"]))

        rotor_mass = inputs["blades_mass"] + inputs["hub_system_mass"]
        nac_mass = inputs["nacelle_mass"]

        # rna mass
        outputs["rotor_mass"] = rotor_mass
        outputs["rna_mass"] = rotor_mass + nac_mass

        # rna cm
        shaft0 = inputs["shaft_start"]
        hub_cm = inputs["hub_system_cm"]
        L_drive = inputs["L_drive"]
        hub_cm = shaft0 + (L_drive + hub_cm) * np.array([Cup * np.cos(tilt), 0.0, np.sin(tilt)])
        outputs["rna_cm"] = (rotor_mass * hub_cm + nac_mass * inputs["nacelle_cm"]) / outputs["rna_mass"]

        # rna I
        hub_I = util.assembleI(util.rotateI(inputs["hub_system_I"], -Cup * tilt, axis="y"))
        blades_I = util.assembleI(util.rotateI(inputs["blades_I"], -Cup * tilt, axis="y"))
        nac_I_TT = util.assembleI(inputs["nacelle_I_TT"])
        rotor_I = blades_I + hub_I

        R = hub_cm
        rotor_I_TT = rotor_I + rotor_mass * (np.dot(R, R) * np.eye(3) - np.outer(R, R))
        outputs["rna_I_TT"] = util.unassembleI(rotor_I_TT + nac_I_TT)


# --------------------------------------------


class DriveDynamics(om.ExplicitComponent):
    """
    Compute equivalent spring constant and damping for the drivetrain system

    Parameters
    ----------
    lss_spring_constant : float, [N*m/rad]
        Equivalent spring constant for the low speed shaft froom T=(G*J/L)*theta
    hss_spring_constant : float, [N*m/rad]
        Equivalent spring constant for the high speed shaft froom T=(G*J/L)*theta
    gear_ratio : float
        overall gearbox ratio
    damping_ratio : float
        Fraction of critical damping value for drivetrain system
    blades_I : numpy array[6], [kg*m**2]
        Mass moments of inertia of all blades about hub center
    hub_system_I : numpy array[6], [kg*m**2]
        Mass moments of inertia of hub system about its CofM


    Returns
    -------
    drivetrain_spring_constant : float, [N*m/rad]
        Equivalent spring constant for the drivetrain system
    drivetrain_damping_coefficient : float, [N*m/rad]
        Equivalent damping coefficient for the drivetrain system
    """

    def setup(self):

        self.add_input("lss_spring_constant", 0.0, units="N*m/rad")
        self.add_input("hss_spring_constant", 0.0, units="N*m/rad")
        self.add_input("gear_ratio", val=1.0)
        self.add_input("damping_ratio", val=0.0)
        self.add_input("blades_I", np.zeros(6), units="kg*m**2")
        self.add_input("hub_system_I", np.zeros(6), units="kg*m**2")

        self.add_output("drivetrain_spring_constant", 0.0, units="N*m/rad")
        self.add_output("drivetrain_damping_coefficient", 0.0, units="N*m*s/rad")

    def compute(self, inputs, outputs):

        # Unpack inputs
        k_lss = inputs["lss_spring_constant"]
        k_hss = inputs["hss_spring_constant"]
        gbr = inputs["gear_ratio"]
        zeta = inputs["damping_ratio"]
        rotor_I = inputs["blades_I"] + inputs["hub_system_I"]

        # springs in series, should be n^2*k1*k2/(k1+n^2*k2)
        # https://www.nrel.gov/docs/fy09osti/41160.pdf
        k_drive = k_lss if gbr == 1.0 else 1.0 / (1 / k_lss + 1 / k_hss / gbr / gbr)
        outputs["drivetrain_spring_constant"] = k_drive

        # Critical damping value
        c_crit = 2.0 * np.sqrt(k_drive * rotor_I[0])
        outputs["drivetrain_damping_coefficient"] = zeta * c_crit


# --------------------------------------------



import openmdao.api as om


class TurbineClass(om.ExplicitComponent):
    """
    Compute velocities based on the chosen turbine class

    Parameters
    ----------
    turbine_class : string
        IEC turbine class
    V_mean_overwrite : float
        overwrite value for mean velocity for using user defined CDFs

    Returns
    -------
    V_mean : float, [m/s]
        IEC mean wind speed for Rayleigh distribution
    V_extreme1 : float, [m/s]
        IEC extreme wind speed at hub height for a 1-year retunr period
    V_extreme50 : float, [m/s]
        IEC extreme wind speed at hub height for a 50-year retunr period

    """

    def setup(self):
        # parameters
        self.add_discrete_input("turbine_class", val="I")
        self.add_input("V_mean_overwrite", val=0.0)

        self.add_output("V_mean", 0.0, units="m/s")
        self.add_output("V_extreme1", 0.0, units="m/s")
        self.add_output("V_extreme50", 0.0, units="m/s")

        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        turbine_class = discrete_inputs["turbine_class"].upper()

        if turbine_class == "I":
            Vref = 50.0
        elif turbine_class == "II":
            Vref = 42.5
        elif turbine_class == "III":
            Vref = 37.5
        elif turbine_class == "IV":
            Vref = 30.0
        else:
            raise ValueError("turbine_class input must be I/II/III/IV")

        if inputs["V_mean_overwrite"] == 0.0:
            outputs["V_mean"] = 0.2 * Vref
        else:
            outputs["V_mean"] = inputs["V_mean_overwrite"]
        outputs["V_extreme1"] = 0.8 * Vref
        outputs["V_extreme50"] = 1.4 * Vref
